#!/usr/bin/python3

import datetime
import json
import os
import random
import re
import signal
import threading
import time

from trex_stl_lib.trex_stl_types import RC_ERR, RC_OK

from trex_stl_lib.utils.text_opts import format_num
from trex_stl_lib.utils.zipmsg import ZippedMsg

import zmq

# import traceback

# from .trex_stl_jsonrpc_client import JsonRpcClient, BatchMessage


# basic async stats class
class CTRexAsyncStats(object):
    def __init__(self):
        self.ref_point = None
        self.current = {}
        self.last_update_ts = datetime.datetime.now()

    def update(self, snapshot):

        # update
        self.last_update_ts = datetime.datetime.now()

        self.current = snapshot

        if self.ref_point is None:
            self.ref_point = self.current

    def clear(self):
        self.ref_point = self.current

    def get(self, field, format=False, suffix=""):

        if field not in self.current:
            return "N/A"

        if not format:
            return self.current[field]
        else:
            return format_num(self.current[field], suffix)

    def get_rel(self, field, format=False, suffix=""):
        if field not in self.current:
            return "N/A"

        if not format:
            return(self.current[field] - self.ref_point[field])
        else:
            return format_num(self.current[field]
                              - self.ref_point[field], suffix)

    # return true if new data has arrived in the past 2 seconds
    def is_online(self):
        delta_ms = (datetime.datetime.now() -
                    self.last_update_ts).total_seconds() * 1000
        return(delta_ms < 2000)


# describes the general stats provided by TRex
class CTRexAsyncStatsGeneral(CTRexAsyncStats):
    def __init__(self):
        super(CTRexAsyncStatsGeneral, self).__init__()


# per port stats
class CTRexAsyncStatsPort(CTRexAsyncStats):
    def __init__(self):
        super(CTRexAsyncStatsPort, self).__init__()

    def get_stream_stats(self, stream_id):
        return None


# stats manager
class CTRexAsyncStatsManager():
    def __init__(self):

        self.general_stats = CTRexAsyncStatsGeneral()
        self.port_stats = {}

    def get_general_stats(self):
        return self.general_stats

    def get_port_stats(self, port_id):

        if not str(port_id) in self.port_stats:
            return None

        return self.port_stats[str(port_id)]

    def update(self, data):
        self.__handle_snapshot(data)

    def __handle_snapshot(self, snapshot):

        general_stats = {}
        port_stats = {}

        # filter the values per port and general
        for key, value in list(snapshot.items()):

            # match a pattern of ports
            m = re.search(r'(.*)\-([0-8])', key)
            if m:

                port_id = m.group(2)
                field_name = m.group(1)

                if port_id not in port_stats:
                    port_stats[port_id] = {}

                port_stats[port_id][field_name] = value

            else:
                # no port match - general stats
                general_stats[key] = value

        # update the general object with the snapshot
        self.general_stats.update(general_stats)

        # update all ports
        for port_id, data in list(port_stats.items()):

            if port_id not in self.port_stats:
                self.port_stats[port_id] = CTRexAsyncStatsPort()

            self.port_stats[port_id].update(data)


class CTRexAsyncClient():
    THREAD_STATE_ACTIVE = 1
    THREAD_STATE_ZOMBIE = 2
    THREAD_STATE_DEAD = 3

    def __init__(self, server, port, stateless_client):

        self.port = port
        self.server = server

        self.stateless_client = stateless_client

        self.event_handler = stateless_client.event_handler
        self.logger = self.stateless_client.logger

        self.raw_snapshot = {}

        self.stats = CTRexAsyncStatsManager()

        self.last_data_recv_ts = 0
        self.async_barrier = None

        self.monitor = AsyncUtil()

        self.connected = False

        self.zipped = ZippedMsg()

        self.t_state = self.THREAD_STATE_DEAD

    # connects the async channel
    def connect(self):

        if self.connected:
            self.disconnect()

        self.tr = "tcp://{0}:{1}".format(self.server, self.port)

        #  Socket to talk to server
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.SUB)

        # before running the thread - mark as active
        self.t_state = self.THREAD_STATE_ACTIVE
        self.t = threading.Thread(target=self._run_safe)

        # kill this thread on exit and don't add it to the join list
        self.t.setDaemon(True)
        self.t.start()

        self.connected = True

        # first barrier - make sure async thread is up
        rc = self.barrier()
        if not rc:
            self.disconnect()
            return rc

        return RC_OK()

    # disconnect
    def disconnect(self):
        if not self.connected:
            return

        # mark for join
        self.t_state = self.THREAD_STATE_DEAD
        self.context.term()
        self.t.join()

        # done
        self.connected = False

    # set the thread as a zombie(in case of server death)

    def set_as_zombie(self):
        self.last_data_recv_ts = None
        self.t_state = self.THREAD_STATE_ZOMBIE

    # return the timeout in seconds for the ZMQ subscriber thread
    def get_timeout_sec(self):
        return 3

    def _run_safe(self):

        # socket must be created on the same thread
        self.socket.setsockopt(zmq.SUBSCRIBE, b'')
        self.socket.setsockopt(zmq.RCVTIMEO, self.get_timeout_sec() * 1000)
        self.socket.connect(self.tr)

        try:
            self._run()
        except Exception as error:
            self.event_handler.on_async_crash()

            # kill all the threads in the group
            os.killpg(os.getppid(), signal.SIGTERM)
            print('General failure: ' + error)

        finally:
            # closing of socket must be from the same thread
            self.socket.close(linger=0)

    # thread function
    def _run(self):

        got_data = False

        self.monitor.reset()

        while self.t_state != self.THREAD_STATE_DEAD:
            try:
                with self.monitor:
                    line = self.socket.recv()

                # last data recv.
                self.last_data_recv_ts = time.time()

                # if thread was marked as zombie
                # - it does nothing besides fetching messages
                if self.t_state == self.THREAD_STATE_ZOMBIE:
                    continue

                self.monitor.on_recv_msg(line)

                # try to decomrpess
                unzipped = self.zipped.decompress(line)
                if unzipped:
                    line = unzipped

                line = line.decode()

                # signal once
                if not got_data:
                    self.event_handler.on_async_alive()
                    got_data = True

            # got a timeout - mark as not alive and retry
            except zmq.Again:
                # signal once
                if got_data:
                    self.event_handler.on_async_timeout(self.get_timeout_sec())
                    got_data = False

                continue

            except zmq.ContextTerminated:
                # outside thread signaled us to exit
                assert(self.t_state != self.THREAD_STATE_ACTIVE)
                break

            msg = json.loads(line)

            name = msg['name']
            data = msg['data']
            type = msg['type']
            baseline = msg.get('baseline', False)

            self.raw_snapshot[name] = data

            self.__dispatch(name, type, data, baseline)

    def get_stats(self):
        return self.stats

    def get_raw_snapshot(self):
        return self.raw_snapshot

    # dispatch the message to the right place
    def __dispatch(self, name, type, data, baseline):

        # stats
        if name == "trex-global":
            self.event_handler.on_async_stats_update(data, baseline)

        # events
        elif name == "trex-event":
            self.event_handler.on_async_event(type, data)

        # barriers
        elif name == "trex-barrier":
            self.handle_async_barrier(type, data)

        elif name == "flow_stats":
            self.event_handler.on_async_rx_stats_event(data, baseline)

        elif name == "latency_stats":
            self.event_handler.on_async_latency_stats_event(data, baseline)

        else:
            pass

    # async barrier handling routine
    def handle_async_barrier(self, type, data):
        if self.async_barrier['key'] == type:
            self.async_barrier['ack'] = True

    # block on barrier for async channel
    def barrier(self, timeout=5, baseline=False):

        # set a random key
        key = random.getrandbits(32)
        self.async_barrier = {'key': key, 'ack': False}

        # expr time
        expr = time.time() + timeout

        while not self.async_barrier['ack']:

            # inject
            rc = self.stateless_client._transmit(
                "publish_now", params={'key': key, 'baseline': baseline})
            if not rc:
                return rc

            # fast loop
            for i in range(0, 100):
                if self.async_barrier['ack']:
                    break
                time.sleep(0.001)

            if time.time() > expr:
                return RC_ERR("*** [subscriber] - timeout "
                              + "- no data flow from server at : " + self.tr)

        return RC_OK()


# a class to measure util. of async subscriber thread
class AsyncUtil(object):

    STATE_SLEEP = 1
    STATE_AWAKE = 2

    def __init__(self):
        self.reset()

    def reset(self):
        self.state = self.STATE_AWAKE
        self.clock = time.time()

        # reset the current interval
        self.interval = {'ts': time.time(), 'total_sleep': 0, 'total_bits': 0}

        # global counters
        self.cpu_util = 0
        self.bps = 0

    def on_recv_msg(self, message):
        self.interval['total_bits'] += len(message) * 8.0

        self._tick()

    def __enter__(self):
        assert(self.state == self.STATE_AWAKE)
        self.state = self.STATE_SLEEP

        self.sleep_start_ts = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        assert(self.state == self.STATE_SLEEP)
        self.state = self.STATE_AWAKE

        # measure total sleep time for interval
        self.interval['total_sleep'] += time.time() - self.sleep_start_ts

        self._tick()

    def _tick(self):
        # how much time did the current interval lasted
        ts = time.time() - self.interval['ts']
        if ts < 1:
            return

        # if tick is in the middle of sleep - add the interval and reset
        if self.state == self.STATE_SLEEP:
            self.interval['total_sleep'] += time.time() - self.sleep_start_ts
            self.sleep_start_ts = time.time()

        # add the interval
        if self.interval['total_sleep'] > 0:
            # calculate
            self.cpu_util = self.cpu_util * 0.75 + \
                (float(ts - self.interval['total_sleep']) / ts) * 0.25
            self.interval['total_sleep'] = 0

        if self.interval['total_bits'] > 0:
            # calculate
            self.bps = self.bps * 0.75 + \
                (self.interval['total_bits'] / ts) * 0.25
            self.interval['total_bits'] = 0

        # reset the interval's clock
        self.interval['ts'] = time.time()

    def get_cpu_util(self):
        self._tick()
        return(self.cpu_util * 100)

    def get_bps(self):
        self._tick()
        return(self.bps)



#!/usr/bin/python3

import os

from scapy.utils import RawPcapReader, RawPcapWriter


def __ts_key(a):
    return float(a[1][0]) + (float(a[1][1]) / 1e6)


def merge_cap_files(pcap_file_list, out_filename, delete_src=False):

    if not all([os.path.exists(f) for f in pcap_file_list]):
        print("failed to merge cap file list...\nnot all files exist\n")
        return

    out_pkts = []
    for src in pcap_file_list:
        pkts = RawPcapReader(src)
        out_pkts += pkts
        if delete_src:
            os.unlink(src)

    # sort by timestamp
    out_pkts = sorted(out_pkts, key=__ts_key)

    writer = RawPcapWriter(out_filename, linktype=1)

    writer._write_header(None)
    for pkt in out_pkts:
        writer._write_packet(pkt[0], sec=pkt[1][0],
                             usec=pkt[1][1], caplen=pkt[1][2], wirelen=None)



#!/usr/bin/python3

"""
ARP service implementation

Description:
    <FILL ME HERE>

How to use:
    <FILL ME HERE>

Author:
  Itay Marom

"""
from collections import defaultdict

from scapy.layers.l2 import ARP, Ether
from trex_stl_lib.services.trex_stl_service import STLService, STLServiceFilter
# from ..trex_stl_types import listify
from trex_stl_lib.trex_stl_vlan import VLAN


class STLServiceFilterARP(STLServiceFilter):
    '''
        Service filter for ARP services
    '''

    def __init__(self):
        self.services = defaultdict(list)

    def add(self, service):
        # forward packets according to the SRC/DST IP
        self.services[(service.src_ip, service.dst_ip,
                       tuple(service.vlan))].append(service)

    def lookup(self, pkt):

        # use scapy to parse
        scapy_pkt = Ether(pkt)

        # not ARP
        if 'ARP' not in scapy_pkt:
            return []

        # forward only ARP response
        if scapy_pkt['ARP'].op != 2:
            return []

        vlans = VLAN.extract(scapy_pkt)

        # ignore VLAN 0 - hash as empty VLAN
        vlans = vlans if vlans != [0] else []

        return self.services.get((scapy_pkt['ARP'].pdst, scapy_pkt['ARP'].psrc, tuple(vlans)), [])

    def get_bpf_filter(self):
        # a simple BPF pattern for ARP(this is not duplicate, it is for QinQ)
        return 'arp or(vlan and arp) or(vlan and arp)'


class STLServiceARP(STLService):
    '''
        ARP service - generate ARP requests
    '''

    def __init__(self, ctx, dst_ip, src_ip='0.0.0.0', vlan=None, timeout_sec=3, verbose_level=STLService.ERROR):

        # init the base object
        super(STLServiceARP, self).__init__(verbose_level)

        self.src_mac = ctx.get_src_mac()
        self.src_ip = src_ip
        self.dst_ip = dst_ip
        self.vlan = VLAN(vlan)
        self.timeout_sec = timeout_sec

        self.record = None

    def get_filter_type(self):
        return STLServiceFilterARP

    def run(self, pipe):
        '''
            Will execute ARP request
        '''

        self.log(
            "ARP: ---> who has '{0}' ? tell '{1}' ".format(self.dst_ip, self.src_ip))

        pkt = Ether(dst="ff:ff:ff:ff:ff:ff")/ARP(psrc=self.src_ip,
                                                 pdst=self.dst_ip, hwsrc=self.src_mac)

        # add VLAN to the packet if needed
        self.vlan.embed(pkt)

        # send the ARP request
        pipe.async_tx_pkt(pkt)

        # wait for RX packet
        pkts = yield pipe.async_wait_for_pkt(time_sec=self.timeout_sec)
        if not pkts:
            self.log("ARP: <--- timeout for '{0}'".format(self.dst_ip))
            self.record = ARPRecord(self.src_ip, self.dst_ip)
            return

        # parse record
        response = Ether(pkts[0]['pkt'])
        self.record = ARPRecord(self.src_ip, self.dst_ip, response)

        self.log(
            "ARP: <--- '{0} is at '{1}'".format(self.record.dst_ip, self.record.dst_mac))

    def get_record(self):
        return self.record


class ARPRecord(object):
    def __init__(self, src_ip='N/A', dst_ip='N/A', scapy_pkt=None):
        self.src_ip = src_ip
        self.dst_ip = dst_ip
        self.dst_mac = None

        if scapy_pkt:
            self.dst_mac = scapy_pkt['ARP'].hwsrc

    def __bool__(self):
        return self.dst_mac is not None

    # __bool__  = __nonzero__

    def __str__(self):
        if self.dst_mac:
            return "Recieved ARP reply from: {0}, hw: {1}".format(self.dst_ip, self.dst_mac)
        else:
            return "Failed to receive ARP response from {0}".format(self.dst_ip)



#!/usr/bin/python3
# base object class for a packet builder


class CTrexPktBuilderInterface(object):

    def compile(self):
        """
        Compiles the packet and VM
        """
        raise Exception("implement me")

    def dump_pkt(self):
        """
        Dumps the packet as a decimal array of bytes(each item x gets value between 0-255)

        :parameters:
            None

        :return:
            + packet representation as array of bytes

        :raises:
            + :exc:`CTRexPktBuilder.EmptyPacketError`, in case packet is empty.

        """

        raise Exception("implement me")

    def get_vm_data(self):
        """
        Dumps the instructions

        :parameters:
            None

        :return:
            + json object of instructions

        """

        raise Exception("implement me")



#!/usr/bin/python3

import zlib
import struct


class ZippedMsg:

    MSG_COMPRESS_THRESHOLD = 256
    MSG_COMPRESS_HEADER_MAGIC = 0xABE85CEA

    def check_threshold(self, msg):
        return len(msg) >= self.MSG_COMPRESS_THRESHOLD

    def compress(self, msg):
        # compress
        compressed = zlib.compress(msg)
        new_msg = struct.pack(
            ">II", self.MSG_COMPRESS_HEADER_MAGIC, len(msg)) + compressed
        return new_msg

    def decompress(self, msg):
        if len(msg) < 8:
            return None

        t = struct.unpack(">II", msg[:8])
        if(t[0] != self.MSG_COMPRESS_HEADER_MAGIC):
            return None

        x = zlib.decompress(msg[8:])
        if len(x) != t[1]:
            return None

        return x

    def is_compressed(self, msg):
        if len(msg) < 8:
            return False

        t = struct.unpack(">II", msg[:8])
        if(t[0] != self.MSG_COMPRESS_HEADER_MAGIC):
            return False

        return True



#!/usr/bin/python3

try:  # Python2
    import queue
    from urllib2 import *
except Exception:  # Python3
    import queue as Queue
    from urllib.request import *
    from urllib.error import *
import threading
import sys
from time import sleep
from pprint import pprint

"""
GAObjClass is a class destined to send Google Analytics Information.

cid - unique number per user.
command - the Event Category rubric appears on site. type: TEXT
action - the Event Action rubric appears on site - type: TEXT
label - the Event Label rubric - type: TEXT
value - the event value metric - type: INTEGER

QUOTAS:
1 single payload - up to 8192Bytes
batched:
A maximum of 20 hits can be specified per request.
The total size of all hit payloads cannot be greater than 16K bytes.
No single hit payload can be greater than 8K bytes.
"""
url_single = 'https://www.google-analytics.com/collect'  # sending single event
url_batched = 'https://www.google-analytics.com/batch'  # sending batched events
url_debug = 'https://www.google-analytics.com/debug/collect'  # verifying hit is valid
# url_conn = 'http://172.217.2.196'  # testing internet connection to this address(google-analytics server)
url_conn = 'http://google.com'


# ..................................................................class GA_ObjClass................................................................
class GA_ObjClass:
    def __init__(self, cid, trackerID, appName, appVer):
        self.cid = cid
        self.trackerID = trackerID
        self.appName = appName
        self.appVer = appVer
        self.payload = ''
        self.payload = GA_ObjClass.generate_payload(self)
        self.size = sys.getsizeof(self.payload)

    def generate_payload(self):
        self.payload += 'v=1&t=event&tid=' + str(self.trackerID)
        self.payload += '&cid=' + str(self.cid)
        self.payload += '&an=' + str(self.appName)
        self.payload += '&av=' + str(self.appVer)
        return self.payload


# ..................................................................class GA_EVENT_ObjClass................................................................
class GA_EVENT_ObjClass(GA_ObjClass):
    def __init__(self, cid, trackerID, command, action, label, value, appName, appVer):
        GA_ObjClass.__init__(self, cid, trackerID, appName, appVer)
        self.command = command
        self.action = action
        self.label = label
        self.value = value
        self.payload = self.generate_payload()
        self.size = sys.getsizeof(self.payload)

    def generate_payload(self):
        self.payload += '&ec=' + str(self.command)
        self.payload += '&ea=' + str(self.action)
        self.payload += '&el=' + str(self.label)
        self.payload += '&ev=' + str(self.value)
        return self.payload


# ..................................................................class GA_EXCEPTION_ObjClass................................................................
# ExceptionFatal - BOOLEAN
class GA_EXCEPTION_ObjClass(GA_ObjClass):
    def __init__(self, cid, trackerID, ExceptionName, ExceptionFatal, appName, appVer):
        GA_ObjClass.__init__(self, cid, trackerID, appName, appVer)
        self.ExceptionName = ExceptionName
        self.ExceptionFatal = ExceptionFatal
        self.payload = self.generate_payload()

    def generate_payload(self):
        self.payload += '&exd=' + str(self.ExceptionName)
        self.payload += '&exf=' + str(self.ExceptionFatal)
        return self.payload


# ..................................................................class GA_TESTING_ObjClass................................................................
class GA_TESTING_ObjClass(GA_ObjClass):
    def __init__(self, cid, trackerID, TRexMode, TestName, SetupName, appName, ActionNumber, appVer, TestType, Mppspc,
                 GoldenMin, GoldenMax):
        GA_ObjClass.__init__(self, cid, trackerID, appName, appVer)
        self.ActionNumber = ActionNumber
        self.TRexMode = TRexMode
        self.TestName = TestName
        self.SetupName = SetupName
        self.TestType = TestType
        self.Mppspc = Mppspc
        self.GoldenMin = GoldenMin
        self.GoldenMax = GoldenMax
        self.payload = self.generate_payload()
        self.size = sys.getsizeof(self.payload)

    def generate_payload(self):
        self.payload += '&ec=TRexTests'
        self.payload += '&ea=' + str(self.ActionNumber)
        self.payload += '&cd2=' + str(self.TRexMode)
        self.payload += '&cd1=' + str(self.TestName)
        self.payload += '&cd3=' + str(self.SetupName)
        self.payload += '&cd4=' + str(self.TestType)
        self.payload += '&cm1=' + str(self.Mppspc)
        self.payload += '&cm2=' + str(self.GoldenMin)
        self.payload += '&cm3=' + str(self.GoldenMax)
        return self.payload

        # .....................................................................class ga_Thread.................................................................


"""

Google analytics thread manager:

will report and empty queue of google analytics items to GA server, every Timeout(parameter given on initialization)
will perform connectivity check every timeout*10 seconds

"""


class ga_Thread(threading.Thread):
    def __init__(self, threadID, gManager):
        threading.Thread.__init__(self)
        self.threadID = threadID
        self.gManager = gManager

    def run(self):
        keepAliveCounter = 0
        # sys.stdout.write('thread started \n')
        # sys.stdout.flush()
        while True:
            if(keepAliveCounter == 10):
                keepAliveCounter = 0
                if(self.gManager.internet_on() == True):
                    self.gManager.connectedToInternet = 1
                else:
                    self.gManager.connectedToInternet = 0
            sleep(self.gManager.Timeout)
            keepAliveCounter += 1
            if not self.gManager.GA_q.empty():
                self.gManager.threadLock.acquire(1)
                #               sys.stdout.write('lock acquired: reporting to GA \n')
                #               sys.stdout.flush()
                if(self.gManager.connectedToInternet == 1):
                    self.gManager.emptyAndReportQ()
                self.gManager.threadLock.release()


# sys.stdout.write('finished \n')
#               sys.stdout.flush()
# .....................................................................class GAmanager.................................................................
"""

Google ID - specify tracker property, example: UA-75220362-2(when the suffix '2' specifies the analytics property profile)

UserID - unique userID, this will differ between users on GA

appName - s string to determine app name

appVer - a string to determine app version

QueueSize - the size of the queue that holds reported items. once the Queue is full:
            on blocking mode:
                will block program until next submission to GA server, which will make new space
            on non-blocking mode:
                will drop new requests

Timout - the timeout the queue uses between data transmissions. Timeout should be shorter than the time it takes to generate 20 events. MIN VALUE = 11 seconds

User Permission - the user must accept data transmission, use this flag as 1/0 flag, when UserPermission=1 allows data collection
 
BlockingMode - set to 1 if you wish every Google Analytic Object will be submitted and processed, with no drops allowed. 
               this will block the running of the program until every item is processed

*** Restriction - Google's restriction for amount of packages being sent per session per second is: 1 event per second, per session. session length is 30min ***
"""


class GAmanager:
    def __init__(self, GoogleID, UserID, appName, appVer, QueueSize, Timeout, UserPermission, BlockingMode):
        self.UserID = UserID
        self.GoogleID = GoogleID
        self.QueueSize = QueueSize
        self.Timeout = Timeout
        self.appName = appName
        self.appVer = appVer
        self.UserPermission = UserPermission
        self.GA_q = queue.Queue(QueueSize)
        self.thread = ga_Thread(UserID, self)
        self.threadLock = threading.Lock()
        self.BlockingMode = BlockingMode
        self.connectedToInternet = 0
        if(self.internet_on() == True):
            #           sys.stdout.write('internet connection active \n')
            #           sys.stdout.flush()
            self.connectedToInternet = 1
        else:
            self.connectedToInternet = 0

    def gaAddAction(self, Event, action, label, value):
        self.gaAddObject(
            GA_EVENT_ObjClass(self.UserID, self.GoogleID, Event, action, label, value, self.appName, self.appVer))

    def gaAddException(self, ExceptionName, ExceptionFatal):
        self.gaAddObject(
            GA_EXCEPTION_ObjClass(self.UserID, self.GoogleID, ExceptionName, ExceptionFatal, self.appName, self.appVer))

    def gaAddObject(self, Object):
        if(self.BlockingMode == 1):
            while(self.GA_q.full()):
                sleep(self.Timeout)
            #               sys.stdout.write('blocking mode=1 \n queue full - sleeping for timeout \n') # within Timout, the thread will empty part of the queue
            #               sys.stdout.flush()
        lockState = self.threadLock.acquire(self.BlockingMode)
        if(lockState == 1):
            #           sys.stdout.write('got lock, adding item \n')
            #           sys.stdout.flush()
            try:
                self.GA_q.put_nowait(Object)
            #               sys.stdout.write('got lock, item added \n')
            #               sys.stdout.flush()
            except queue.Full:
                #               sys.stdout.write('Queue full \n')
                #               sys.stdout.flush()
                pass
            self.threadLock.release()

    def emptyQueueToList(self, obj_list):
        items = 0
        while((not self.GA_q.empty()) and(items < 20)):
            obj_list.append(self.GA_q.get_nowait().payload)
            items += 1
        #           print items
        return obj_list

    def reportBatched(self, batched):
        req = Request(url_batched, data=batched.encode('ascii'))
        urlopen(req)

    #       pprint(r.json())

    def emptyAndReportQ(self):
        obj_list = []
        obj_list = self.emptyQueueToList(obj_list)
        if(len(obj_list) == 0):
            return
        batched = '\n'.join(obj_list)
        #       print sys.getsizeof(batched)
        #       print batched # - for debug
        self.reportBatched(batched)

    def printSelf(self):
        print('remaining in queue:')
        while not self.GA_q.empty():
            obj = self.GA_q.get_nowait()
            print((obj.payload))

    def internet_on(self):
        try:
            urlopen(url_conn, timeout=10)
            return True
        except URLError as err:
            pass
        return False

    def activate(self):
        if(self.UserPermission == 1):
            self.thread.start()


# .....................................................................class GAmanager_Regression.................................................................
"""
            *-*-*-*-Google Analytics Regression Manager-*-*-*-*
            attributes:
GoogleID - the tracker ID that Google uses in order to track the activity of a property. for regression use: 'UA-75220362-4'
AnalyticsUserID - text value  - used by Google to differ between 2 users sending data.(will not be presented on reports). use only as a way to differ between different users
TRexMode - text - will be presented on analysis. put here TRexMode
appName - text - will be presented on analysis. put here appName as string describing app name
appVer - text - will be presented on analysis. put here the appVer
QueueSize - integer - determines the queue size. the queue will hold pending request before submission. RECOMMENDED VALUE: 20
Timeout - integer(seconds) - the timeout in seconds between automated reports when activating reporting thread
UserPermission - boolean(1/0) - required in order to send packets, should be 1.
BlockingMode - boolean(1/0) - required when each tracked event is critical and program should halt until the event is reported
SetupName - text - will be presented on analysis. put here setup name as string.
"""


class GAmanager_Regression(GAmanager):
    def __init__(self, GoogleID, AnalyticsUserID, appName, appVer,
                 QueueSize, Timeout, UserPermission, BlockingMode):
        GAmanager.__init__(self, GoogleID, AnalyticsUserID, appName, appVer,
                           QueueSize, Timeout, UserPermission, BlockingMode)
        self.GoogleID = GoogleID
        self.AnalyticsUserID = AnalyticsUserID

    def gaAddTestQuery(self, TestName, TRexMode, SetupName, ActionNumber, TestType, Mppspc, GoldenMin, GoldenMax):
        self.gaAddObject(
            GA_TESTING_ObjClass(self.AnalyticsUserID, self.GoogleID, TRexMode, TestName, SetupName, self.appName,
                                ActionNumber, self.appVer, TestType, Mppspc, GoldenMin, GoldenMax))



#!/usr/bin/python3

# from trex_stl_lib.api import *
from .trex_stl_service import STLService, STLServiceFilter
from .trex_stl_service_int import STLServiceCtx, simpy, TXBuffer
import time
from collections import deque
from scapy.all import *
from scapy.contrib.capwap import *
from trex_openssl import *
import threading
import sys
import time
import base64

'''
FSMs for AP:
   * Discover WLC
   * Establish DTLS session
   * Join WLC
   * Add client(station)
   * Shutdown DTLS session
   * Maintenance(arp, ping, capwap echo request, fetches rx and dispatches to rx_buffer of APs)
'''


class STLServiceBufferedCtx(STLServiceCtx):
    ''' Same as parent, but does not use capture to get packets, uses AP's rx_buffer '''

    def _run(self, services):
        self._reset()
        self._add(services)
        if len(self.filters) > 1:
            raise Exception(
                'Services here should have one common filter per AP')

        self.filter = list(self.filters.values())[0]['inst']
        if not hasattr(self.filter, 'services_per_ap'):
            raise Exception('Services here should have filter with attribute services_per_ap, got %s, type: %s' % (
                self.filter, type(self.filter)))

        # create an environment
        self.env = simpy.rt.RealtimeEnvironment(factor=1, strict=False)
        self.tx_buffer = TXBuffer(self.env, self.client, self.port)

        # create processes
        for service in self.services:
            pipe = self._pipe()
            self.services[service]['pipe'] = pipe
            p = self.env.process(service.run(pipe))
            self._on_process_create(p)

        try:
            tick_process = self.env.process(self._tick_process())
            self.env.run(until=tick_process)
        finally:
            self._reset()

    def _tick_process(self):
        while True:

            self.tx_buffer.send_all()

            for ap, services in list(self.filter.services_per_ap.items()):
                for _ in range(len(ap.rx_buffer)):
                    try:
                        scapy_pkt = ap.rx_buffer.popleft()
                    except IndexError:
                        break
                    for service in services:
                        self.services[service]['pipe']._on_rx_pkt(
                            scapy_pkt, None)

            # if no other process exists - exit
            if self.active_services == 0:
                return
            else:
                # backoff
                yield self.env.timeout(0.05)


'''
Just assign services to AP, it will get packets from AP's rx_buffer
'''


class STLServiceFilterPerAp(STLServiceFilter):
    def __init__(self):
        self.services_per_ap = {}

    def add(self, service):
        if service.ap in self.services_per_ap:
            self.services_per_ap[service.ap].append(service)
        else:
            self.services_per_ap[service.ap] = [service]


'''
Used to fetch RX packets for all APs
Decrypts them if possible
Sends echo request(control keep alive)
Answers to async config changes
Does not use SimPy
'''


class STLServiceApBgMaintenance:
    bpf_filter = ('arp or(ip and(icmp or udp src port 5246 or '  # arp, ping, capwap control
                  # capwap data keep-alive
                  + '(udp src port 5247 and(udp[11] & 8 == 8 or '
                  # client assoc. resp
                  + 'udp[16:2] == 16 or '
                  # client arp
                  + 'udp[48:2] == 2054 or '
                  + '(udp[48:2] == 2048 and udp[59] == 1)))))')      # client ping

    ARP_ETHTYPE = b'\x08\x06'
    IP_ETHTYPE = b'\x08\x00'
    ICMP_PROTO = b'\x01'
    UDP_PROTO = b'\x11'
    CAPWAP_CTRL_PORT = b'\x14\x7e'
    CAPWAP_DATA_PORT = b'\x14\x7f'
    WLAN_ASSOC_RESP = b'\x00\x10'
    ARP_REQ = b'\x00\x01'
    ARP_REP = b'\x00\x02'
    ICMP_REQ = b'\x08'

    def __init__(self, ap_mngr, port_id):
        self.ap_mngr = ap_mngr
        self.port_id = port_id
        self.ap_per_ip = {}
        self.client_per_ip = {}
        self.bg_client = self.ap_mngr.bg_client
        self.port = self.bg_client.ports[port_id]
        self.capture_id = None
        self.bg_thread = None

################
#     API      #
################

    def run(self):
        self.bg_thread = threading.Thread(target=self.main_loop_wrapper)
        self.bg_thread.name = 'BG Thread(port %s)' % self.port_id
        self.bg_thread.daemon = True
        self.bg_thread.start()

    def is_running(self):
        return self.bg_thread and self.bg_thread.is_alive()

    def stop(self):
        try:
            self.bg_client.stop_capture(self.capture_id)
        except:
            pass
        self.capture_id = None


##################
#    INTERNAL    #
##################

    def log(self, msg, level=LoggerApi.VERBOSE_REGULAR):
        if not msg.startswith('(WLC) '):
            msg = '(WLC) %s' % msg
        self.ap_mngr.trex_client.logger.async_log('\n' + bold(msg), level)

    def err(self, msg):
        self.log(msg, LoggerApi.VERBOSE_QUIET)

    def fatal(self, msg):
        self.err(msg)
        self.stop()

    def send(self, pkts):
        assert type(pkts) is list
        push_pkts = [{'binary': base64.b64encode(bytes(p) if isinstance(p, Ether) else p).decode(),
                      'use_port_dst_mac': False,
                      'use_port_src_mac': False} for p in pkts]
        rc = self.port.push_packets(push_pkts, False)
        # if not rc:
        #    self.err(rc.err())

    def recv(self):
        pkts = []
        self.bg_client.fetch_capture_packets(self.capture_id, pkts, 1000)
        if len(pkts) > 995:
            self.err('Too much packets in rx queue(%s)' % len(pkts))
        return pkts

    def shutdown_ap(self, ap):
        try:
            for client in ap.clients:
                client.disconnect()
            if ap.is_dtls_established():
                with ap.ssl_lock:
                    libssl.SSL_shutdown(ap.ssl)
                tx_pkt = ap.wrap_capwap_pkt(b'\1\0\0\0' + ap.ssl_read())
                self.send([tx_pkt])
        finally:
            ap.reset_vars()

    def main_loop_wrapper(self):
        err_msg = ''
        try:
            self.capture_id = self.bg_client.start_capture(
                rx_ports=self.port_id, bpf_filter=self.bpf_filter, limit=5000)['id']
            self.main_loop()
        except KeyboardInterrupt:
            pass
        except Exception as e:
            if self.capture_id:  # if no id -> got stop()
                if not isinstance(e, STLError):
                    import traceback
                    traceback.print_exc()
                err_msg = '(Exception: %s)' % e
        finally:
            try:
                self.bg_client.stop_capture(self.capture_id)
            except:
                pass
            if self.port_id in self.ap_mngr.service_ctx:
                if self.ap_per_ip:
                    self.err('Background thread on port %s died%s. Disconnecting APs.' % (
                        self.port_id, err_msg))
                else:
                    self.err('Background thread on port %s died%s.' %
                             (self.port_id, err_msg))
                for ap in list(self.ap_per_ip.values()):
                    self.shutdown_ap(ap)

    def handle_ap_arp(self, rx_bytes, send_pkts):
        ip = rx_bytes[38:42]
        if ip not in self.ap_per_ip:  # check IP
            return
        ap = self.ap_per_ip[ip]
        dst_mac = rx_bytes[:6]
        if dst_mac not in(b'\xff\xff\xff\xff\xff\xff', ap.mac_src):  # check MAC
            ap.err('Bad MAC(%s), although IP of AP(%s)' %
                   (str2mac(dst_mac), str2ip(ip)))
            return
        rx_pkt = Ether(rx_bytes)
        if ap.is_debug:
            ap.debug('AP %s got ARP' % ap.name)
            rx_pkt.show2()

        rx_arp = rx_pkt[ARP]
        if rx_arp.op == 1:  # 'who-has'
            if rx_arp.psrc == rx_arp.pdst:    # GARP
                return
            if rx_arp.pdst == ap.ip_hum:      # arp request to AP
                #print 'Arp to AP'
                tx_pkt = rx_pkt.copy()
                tx_pkt.dst = tx_pkt.src
                tx_pkt.src = ap.mac_src
                tx_pkt[ARP].op = 'is-at'
                tx_pkt[ARP].hwdst = rx_arp.hwsrc
                tx_pkt[ARP].pdst = rx_arp.psrc
                tx_pkt[ARP].hwsrc = ap.mac_src
                tx_pkt[ARP].psrc = ap.ip_hum
                # tx_pkt.show2()
                send_pkts.append(tx_pkt)

        elif rx_arp.op == 2:  # 'is-at'
            ap.rx_buffer.append(rx_pkt)

    def handle_ap_icmp(self, rx_bytes, ap, send_pkts):
        rx_pkt = Ether(rx_bytes)
        icmp_pkt = rx_pkt[ICMP]
        if icmp_pkt.type == 8:  # echo-request
            #print 'Ping to AP!'
            # rx_pkt.show2()
            if rx_pkt[IP].dst == ap.ip_hum:  # ping to AP
                tx_pkt = rx_pkt.copy()
                tx_pkt.src, tx_pkt.dst = tx_pkt.dst, tx_pkt.src
                tx_pkt[IP].src, tx_pkt[IP].dst = tx_pkt[IP].dst, tx_pkt[IP].src
                tx_pkt[ICMP].type = 'echo-reply'
                del tx_pkt[ICMP].chksum
                # tx_pkt.show2()
                send_pkts.append(tx_pkt)

        elif icmp_pkt.type == 0:  # echo-reply
            ap.rx_buffer.append(rx_pkt)

    def process_capwap_ctrl(self, rx_bytes, ap, send_pkts):
        ap.info('Got CAPWAP CTRL at AP %s' % ap.name)
        rx_pkt = CAPWAP_CTRL(rx_bytes[42:])
        if ap.is_debug:
            rx_pkt.show2()
        capwap_assemble = ap.capwap_assemble
        if not ap.is_dtls_established():
            if CAPWAP_Control_Header in rx_pkt and rx_pkt.control_header.type == 2:
                ap.mac_dst = str2mac(rx_bytes[6:12])
                ap.ip_dst = str2ip(rx_bytes[26:30])
            ap.rx_buffer.append(rx_pkt)
            return

        if rx_pkt.is_dtls():  # decrypt
            dtls_buf = bytes(rx_pkt.payload)
            if(dtls_buf[0:1] == '\x15'):
                ap.is_connected = False
                self.err("Server sent DTLS alert to AP '%s'." % ap.name)
            rx_pkt_buf = ap.decrypt(dtls_buf)
            if not rx_pkt_buf:
                return
            # definitely not CAPWAP... should we debug it?
            if rx_pkt_buf[0:1] not in(b'\0', b'\1'):
                ap.debug('Not CAPWAP, skipping: %s' % hex(rx_pkt_buf))
                return
            rx_pkt = CAPWAP_CTRL(rx_pkt_buf)
            ap.last_recv_ts = time.time()

        if ap.is_debug:
            rx_pkt.show2()
        if rx_pkt.header.is_fragment():
            if capwap_assemble:
                assert ap.capwap_assemble[
                    'header'].fragment_id == rx_pkt.header.fragment_id, 'Got CAPWAP fragments with out of order(different fragment ids)'
                control_str = bytes(rx_pkt[CAPWAP_Control_Header_Fragment])
                if rx_pkt.header.fragment_offset * 8 != len(capwap_assemble['buf']):
                    self.err('Fragment offset and data length mismatch')
                    capwap_assemble.clear()
                    return
                # if rx_pkt.header.fragment_offset * 8 > len(capwap_assemble['buf']):
                #    print('Fragment offset: %s, data so far length: %s(not enough data)' %(rx_pkt.header.fragment_offset, len(capwap_assemble['buf'])))
                # elif rx_pkt.header.fragment_offset * 8 < len(capwap_assemble['buf']):
                #    capwap_assemble['buf'] = capwap_assemble['buf'][:rx_pkt.header.fragment_offset * 8]

                capwap_assemble['buf'] += control_str

                if rx_pkt.is_last_fragment():
                    capwap_assemble['assembled'] = CAPWAP_CTRL(
                        header=capwap_assemble['header'],
                        control_header=CAPWAP_Control_Header(
                            capwap_assemble['buf'])
                    )

            else:
                if rx_pkt.is_last_fragment():
                    self.err(
                        'Got CAPWAP first fragment that is also last fragment!')
                    return
                if rx_pkt.header.fragment_offset != 0:
                    rx_pkt.show2()
                    self.err(
                        'Got out of order CAPWAP fragment, does not start with zero offset')
                    return
                capwap_assemble['header'] = rx_pkt.header
                capwap_assemble['header'].flags &= ~0b11000
                capwap_assemble['buf'] = bytes(
                    rx_pkt[CAPWAP_Control_Header_Fragment])
                capwap_assemble['ap'] = ap

        elif capwap_assemble:
            self.err('Got not fragment in middle of assemble of fragments(OOO).')
            capwap_assemble.clear()
        else:
            assert CAPWAP_Control_Header in rx_pkt, rx_pkt.command()
            capwap_assemble['assembled'] = rx_pkt

        rx_pkt = capwap_assemble.get('assembled')
        if not rx_pkt:
            return
        capwap_assemble.clear()

        if ap.is_debug and CAPWAP_CTRL in rx_pkt:
            rx_pkt[CAPWAP_CTRL].show2()
        if CAPWAP_Control_Header not in rx_pkt:
            return
        if rx_pkt.control_header.type == 7:  # Configuration Update Request
            # rx_pkt.show2()
            CAPWAP_PKTS.parse_message_elements(
                rx_pkt, ap, self.ap_mngr)  # get info from incoming packet
            seq = rx_pkt.control_header.sequence_number
            tx_pkt = ap.get_config_update_capwap(seq)
            if ap.is_debug:
                CAPWAP_CTRL(tx_pkt.value).show2()
            send_pkts.append(ap.wrap_capwap_pkt(
                b'\1\0\0\0' + ap.encrypt(tx_pkt)))
        elif rx_pkt.control_header.type == 14:  # Echo Response
            ap.echo_resp_timer = None
        elif rx_pkt.control_header.type == 17:  # Reset Request
            self.err('AP %s got Reset request, shutting down' % ap.name)
            #send_pkts.append(ap.wrap_capwap_pkt(b'\1\0\0\0' + ap.encrypt(tx_pkt)))
            self.shutdown_ap(ap)
        else:
            # rx_pkt.show2()
            ap.rx_buffer.append(rx_pkt)

    def handle_client_arp(self, dot11_bytes, ap, send_pkts):
        ip = dot11_bytes[58:62]
        client = self.client_per_ip.get(ip)
        if not client:
            return
        if client.ap is not ap:
            self.err('Got ARP to client %s via wrong AP(%s)' %
                     (client.ip, ap.name))
            return

        if dot11_bytes[40:42] == self.ARP_REQ:  # 'who-has'
            if dot11_bytes[48:52] == dot11_bytes[58:62]:  # GARP
                return
            tx_pkt = ap.wrap_pkt_by_wlan(client, ap.get_arp_pkt(
                client, 'is-at', dot11_bytes[42:48], str2ip(dot11_bytes[48:52])))
            send_pkts.append(tx_pkt)

        elif dot11_bytes[40:42] == self.ARP_REP:  # 'is-at'
            ap.rx_buffer.append(Dot11_swapped(dot11_bytes))

    def handle_client_icmp(self, dot11_bytes, ap, send_pkts):
        ip = dot11_bytes[50:54]
        client = self.client_per_ip.get(ip)
        if not client:
            return
        if client.ap is not ap:
            self.err('Got ARP to client %s via wrong AP(%s)' %
                     (client.ip, ap.name))
            return

        if dot11_bytes[54:55] == self.ICMP_REQ:
            rx_pkt = Dot11_swapped(dot11_bytes)
            tx_pkt = Ether(src=client.mac, dst=rx_pkt.addr3) / \
                rx_pkt[IP].copy()
            tx_pkt[IP].src, tx_pkt[IP].dst = tx_pkt[IP].dst, tx_pkt[IP].src
            tx_pkt[ICMP].type = 'echo-reply'
            del tx_pkt[ICMP].chksum
            send_pkts.append(ap.wrap_pkt_by_wlan(client, tx_pkt))

    def main_loop(self):
        echo_send_timer = PassiveTimer(1)
        send_pkts = []
        while True:
            now_time = time.time()
            self.send(send_pkts)
            send_pkts = []
            resps = self.recv()

            try:
                # update only if required
                if not self.ap_mngr.service_ctx[self.port_id]['synced']:
                    with self.ap_mngr.bg_lock:
                        aps = list(self.ap_mngr.aps)
                        clients = list(self.ap_mngr.clients)
                        self.ap_mngr.service_ctx[self.port_id]['synced'] = True
                    self.ap_per_ip = dict(
                        [(ap.ip_src, ap) for ap in aps if ap.port_id == self.port_id])
                    self.client_per_ip = dict(
                        [(client.ip_bytes, client) for client in clients])
            except KeyError as e:
                if self.port_id not in self.ap_mngr.service_ctx:
                    return

            if echo_send_timer.has_expired():
                echo_send_timer = PassiveTimer(0.5)

                for ap in list(self.ap_per_ip.values()):
                    if ap.is_connected:
                        if ap.echo_resp_timer and ap.echo_resp_timer.has_expired():  # retry echo
                            if ap.echo_resp_retry > 0:
                                ap.echo_resp_timeout *= 2
                                ap.echo_resp_timer = PassiveTimer(
                                    ap.echo_resp_timeout)
                                ap.echo_resp_retry -= 1
                                tx_pkt = ap.get_echo_capwap()
                                send_pkts.append(
                                    ap.get_echo_wrap(ap.encrypt(tx_pkt)))
                            else:
                                self.err(
                                    "Timeout in echo response for AP '%s', disconnecting" % ap.name)
                                self.shutdown_ap(ap)

                for ap in list(self.ap_per_ip.values()):
                    if ap.is_connected:
                        if time.time() > ap.last_echo_req_ts + ap.echo_req_interval:  # new echoes
                            tx_pkt = ap.get_echo_capwap()
                            ap.last_echo_req_ts = time.time()
                            ap.echo_resp_timeout = ap.capwap_RetransmitInterval
                            ap.echo_resp_timer = PassiveTimer(
                                ap.echo_resp_timeout)
                            ap.echo_resp_retry = ap.capwap_MaxRetransmit
                            send_pkts.append(
                                ap.get_echo_wrap(ap.encrypt(tx_pkt)))
                    if len(send_pkts) > 200:
                        break

            if not resps and not send_pkts:
                time.sleep(0.01)
                continue

            for resp in resps:
                rx_bytes = resp['binary']
                dst_mac = rx_bytes[:7]
                ether_type = rx_bytes[12:14]

                if ether_type == self.ARP_ETHTYPE:
                    self.handle_ap_arp(rx_bytes, send_pkts)

                elif ether_type == self.IP_ETHTYPE:
                    ip = rx_bytes[30:34]
                    if ip not in self.ap_per_ip:  # check IP
                        continue
                    ap = self.ap_per_ip[ip]
                    dst_mac = rx_bytes[:6]
                    if dst_mac not in('\xff\xff\xff\xff\xff\xff', ap.mac_src):  # check MAC
                        ap.err('Bad MAC(%s), although IP of AP(%s)' %
                               (str2mac(dst_mac), str2ip(ip)))
                        continue

                    ip_proto = rx_bytes[23:24]

                    if ip_proto == self.ICMP_PROTO:
                        self.handle_ap_icmp(rx_bytes, ap, send_pkts)

                    elif ip_proto == self.UDP_PROTO:
                        udp_port_str = rx_bytes[36:38]
                        if udp_port_str != ap.udp_port_str:  # check UDP port
                            ap.err('Bad UDP port(%s), although IP of AP(%s)' %
                                   (str2int(udp_port), str2ip(ip)))
                            continue
                        udp_src = rx_bytes[34:36]

                        if udp_src == self.CAPWAP_CTRL_PORT:
                            self.process_capwap_ctrl(rx_bytes, ap, send_pkts)

                        elif udp_src == self.CAPWAP_DATA_PORT:
                            if ord(rx_bytes[45:46]) & 0b1000:  # CAPWAP Data Keep-alive
                                ap.rx_buffer.append(CAPWAP_DATA(rx_bytes[42:]))
                                continue

                            dot11_offset = 42 + \
                                ((ord(rx_bytes[43:44]) & 0b11111000) >> 1)
                            dot11_bytes = rx_bytes[dot11_offset:]

                            # Client assoc. response
                            if dot11_bytes[:2] == self.WLAN_ASSOC_RESP:
                                ap.rx_buffer.append(Dot11_swapped(dot11_bytes))

                            elif dot11_bytes[32:34] == self.ARP_ETHTYPE:
                                self.handle_client_arp(
                                    dot11_bytes, ap, send_pkts)

                            elif dot11_bytes[32:34] == self.IP_ETHTYPE and dot11_bytes[43:44] == self.ICMP_PROTO:
                                self.handle_client_icmp(
                                    dot11_bytes, ap, send_pkts)


class STLServiceAp(STLService):
    requires_dtls = True

    def __init__(self, ap, verbose_level=STLService.WARN):
        STLService.__init__(self, verbose_level)
        self.ap = ap
        self.name = '%s of %s' % (self.__class__.__name__, ap.name)

    def get_filter_type(self):
        return STLServiceFilterPerAp

    @staticmethod
    def wrap_capwap_static(ap, dst_mac, dst_ip, dst_port, pkt):
        assert ap and dst_mac and dst_ip and dst_port
        return(
            Ether(src=ap.mac_src, dst=dst_mac) /
            IP(src=ap.ip_hum, dst=dst_ip) /
            UDP(sport=ap.udp_port, dport=dst_port) /
            pkt)

    def timeout(self):
        self.ap.warn('Timeout in FSM %s' % self.name)

    def log(self, msg):
        self.ap.info(msg)

    def err(self, msg):
        self.ap.err('Error in FSM %s: %s' % (self.name, msg))

    def run(self, pipe):
        if self.requires_dtls and not self.ap.is_dtls_established():
            self.err('DTLS is not established for AP %s' % self.ap.name)
            return
        self.ap.info('Starting FSM %s' % self.name)
        run_gen = self.run_with_buffer()
        send_data = None
        while True:
            if self.requires_dtls and not self.ap.is_dtls_established():
                self.log('DTLS session got closed for AP %s, exiting FSM' %
                         self.ap.name)
                break
            try:
                action = run_gen.send(send_data)
            except StopIteration:
                action = 'done'
            if type(action) is tuple and len(action) == 2:
                action, val = action

            if action == 'get':
                send_data = None
                resp = yield pipe.async_wait_for_pkt(time_sec=val, limit=1)
                if resp:
                    send_data = resp[0]['pkt']

            elif action == 'put':
                if type(val) is list:
                    pipe.async_tx_pkt([PacketBuffer(v) for v in val])
                else:
                    pipe.async_tx_pkt(PacketBuffer(val))

            elif action == 'sleep':
                pipe.async_wait(val)

            elif action == 'done':
                self.log('Finished successfully FSM %s' % self.name)
                break

            elif action == 'err':
                self.err(val)
                break

            elif action == 'time':
                self.timeout()
                break

            else:
                raise Exception('Incorrect action in FSM %s: %s' %
                                (self.name, action))


def hex(buf, delimiter=' '):
    if not buf:
        return 'Empty buffer'
    return delimiter.join(['%02x' % (c if type(c) is int else ord(c)) for c in buf])

################ FSMs ##################


class STLServiceApDiscoverWLC(STLServiceAp):
    requires_dtls = False

    def run_with_buffer(self):
        RetransmitInterval = self.ap.capwap_RetransmitInterval
        for _ in range(max(2, self.ap.capwap_MaxRetransmit)):
            RetransmitInterval *= 2
            discovery_pkt = self.ap.wrap_capwap_pkt(CAPWAP_PKTS.discovery(
                self.ap), dst_mac='ff:ff:ff:ff:ff:ff', dst_ip='255.255.255.255')
            yield('put', discovery_pkt)
            timer = PassiveTimer(RetransmitInterval)
            while not timer.has_expired():
                rx_pkt = yield('get', RetransmitInterval)
                if not rx_pkt:
                    continue
                if not rx_pkt or CAPWAP_Control_Header not in rx_pkt or rx_pkt.control_header.type != 2:
                    continue
                if self.ap.is_debug:
                    rx_pkt.show2()

                for ME in rx_pkt.control_header.message_elements:
                    if ME.type == 10:
                        self.ap.ip_dst = ME.ip_address  # give higher priority to ME
                    if ME.type == 1:
                        if ME.max_wtps <= ME.active_wtps:
                            yield('err', 'WLC reached maximal number of APs connected(%s)!' % ME.max_wtps)
                self.log('Got discovery response from %s' % self.ap.ip_dst)
                return

        yield 'time'


class STLServiceApEstablishDTLS(STLServiceAp):
    requires_dtls = False
    aps_by_ssl = {}

    @staticmethod
    def openssl_callback(ssl, where, ret):
        pipe = STLServiceApEstablishDTLS.aps_by_ssl[ssl]['pipe']
        ap = STLServiceApEstablishDTLS.aps_by_ssl[ssl]['ap']

        if libcrypto.BIO_ctrl_pending(ap.out_bio):
            ssl_data = ap.ssl_read()
            if ssl_data:
                pkt = STLServiceAp.wrap_capwap_static(
                    ap, ap.mac_dst, ap.ip_dst, 5246, b'\1\0\0\0' + ssl_data)
                pipe.async_tx_pkt(PacketBuffer(pkt))

        return 0

    ssl_info_callback_type = CFUNCTYPE(c_int, c_void_p, c_int, c_int)
    ssl_info_callback_func = ssl_info_callback_type(openssl_callback.__func__)

    def run(self, pipe):
        assert self.ap.ssl and(self.ap.ssl not in self.aps_by_ssl)
        #assert not self.ap.is_dtls_established(), 'AP %s has already established DTLS connection!' % self.ap.name
        self.aps_by_ssl[self.ap.ssl] = {'ap': self.ap, 'pipe': pipe}
        with self.ap.ssl_lock:
            libssl.SSL_clear(self.ap.ssl)
            libssl.SSL_set_info_callback(
                self.ap.ssl, self.ssl_info_callback_func)  # set ssl callback
            libssl.SSL_do_handshake(self.ap.ssl)
        try:
            timer = PassiveTimer(5)
            self.ap.info('Start handshake')
            while not timer.has_expired():
                if self.ap.is_handshake_done():
                    return
                if self.ap.is_dtls_closed():
                    self.err('DTLS session got closed for ap %s' %
                             self.ap.name)
                    return

                resps = yield pipe.async_wait_for_pkt(time_sec=1, limit=1)
                if not resps:
                    continue
                pkt = resps[0]['pkt']
                if isinstance(pkt, CAPWAP_CTRL) and pkt.is_dtls():
                    dtls_buf = bytes(pkt.payload)
                    self.ap.decrypt(dtls_buf)

            self.timeout()

        finally:
            with self.ap.ssl_lock:
                libssl.SSL_set_info_callback(
                    self.ap.ssl, None)  # remove ssl callback
            if self.ap.ssl in self.aps_by_ssl:
                del self.aps_by_ssl[self.ap.ssl]


class STLServiceApEncryptedControl(STLServiceAp):

    def control_round_trip(self, tx_pkt, expected_response_type):
        RetransmitInterval = self.ap.capwap_RetransmitInterval
        if isinstance(tx_pkt, Packet) and self.ap.is_debug:
            tx_pkt.show2()

        for _ in range(self.ap.capwap_MaxRetransmit):
            RetransmitInterval *= 2
            tx_pkt = self.ap.wrap_capwap_pkt(
                b'\1\0\0\0' + self.ap.encrypt(tx_pkt))
            yield('put', tx_pkt)

            timer = PassiveTimer(RetransmitInterval)
            while not timer.has_expired():
                rx_pkt = yield('get', RetransmitInterval)
                if not rx_pkt:
                    continue

                if CAPWAP_Control_Header not in rx_pkt:
                    continue
                ctrl_header = rx_pkt[CAPWAP_Control_Header]
                if ctrl_header.type != expected_response_type:
                    continue

                result_code = None
                for ME in ctrl_header.message_elements:
                    if ME.type == 33:
                        result_code = ME.result_code
                        break
                if result_code not in(None, 0, 2):
                    yield('err', 'Not successful result %s - %s.' % (result_code, capwap_result_codes.get(result_code, 'Unknown')))

                if self.ap.is_debug:
                    rx_pkt.show2()

                yield('resp', rx_pkt)

        yield 'time'


class STLServiceApJoinWLC(STLServiceApEncryptedControl):
    def run_with_buffer(self):
        self.log('Sending Join Request')
        ctrl_gen = self.control_round_trip(CAPWAP_PKTS.join(self.ap), 4)
        send_data = None
        while True:
            action = ctrl_gen.send(send_data)
            if type(action) is tuple and action[0] == 'resp':
                resp = action[1]
                ctrl_gen.close()
                break
            else:
                send_data = yield action

        for ME in resp[CAPWAP_Control_Header].message_elements:
            if ME.type == 4:
                self.ap.wlc_name = ME.ac_name
        self.log('Got Join Response')

        self.log('Sending Configuration Status Request')
        ctrl_gen = self.control_round_trip(
            CAPWAP_PKTS.conf_status_req(self.ap), 6)
        send_data = None
        while True:
            action = ctrl_gen.send(send_data)
            if type(action) is tuple and action[0] == 'resp':
                resp = action[1]
                ctrl_gen.close()
                break
            else:
                send_data = yield action
        self.log('Got Configuration Status Response')

        self.log('Sending Change State Event Request')
        ctrl_gen = self.control_round_trip(
            CAPWAP_PKTS.change_state(self.ap), 12)
        send_data = None
        while True:
            action = ctrl_gen.send(send_data)
            if type(action) is tuple and action[0] == 'resp':
                resp = action[1]
                ctrl_gen.close()
                break
            else:
                send_data = yield action
        self.log('Got Change State Event Response')

        self.log('Going to ack all config updates and try to get SSID')
        while self.ap.last_recv_ts + 5 > time.time():  # ack all config updates in BG thread
            if self.ap.SSID:
                break
            if self.ap.is_dtls_closed():
                return
            yield('sleep', 0.5)

        if not self.ap.SSID:
            yield('err', 'Did not get SSID from WLC!')

        self.log('Sending Keep-alive.')
        RetransmitInterval = self.ap.capwap_RetransmitInterval
        for _ in range(self.ap.capwap_MaxRetransmit):
            RetransmitInterval *= 2
            tx_pkt = self.ap.wrap_capwap_pkt(
                CAPWAP_PKTS.keep_alive(self.ap), dst_port=5247)
            if self.ap.is_debug:
                Ether(tx_pkt).show2()
            yield('put', tx_pkt)
            timer = PassiveTimer(RetransmitInterval)
            while not timer.has_expired():
                rx_pkt = yield('get', RetransmitInterval)
                if rx_pkt and isinstance(rx_pkt, CAPWAP_DATA) and rx_pkt.is_keepalive():
                    self.log('Received Keep-alive response.')
                    self.ap.last_echo_req_ts = time.time()
                    self.ap.is_connected = True
                    return
        yield 'time'


class STLServiceApAddClients(STLServiceAp):
    def __init__(self, ap, clients, verbose_level=STLService.WARN):
        STLServiceAp.__init__(self, ap, verbose_level)
        assert type(clients) is list
        assert all([hasattr(c, 'mac') and hasattr(c, 'ip')
                    for c in clients]), 'Clients should have attributes mac and ip'
        self.clients = clients

    def run_with_buffer(self):
        self.log('Sending Association requests.')
        need_assoc_resp_clients = dict(
            [(client.mac, client) for client in self.clients])

        RetransmitInterval = self.ap.capwap_RetransmitInterval
        for _ in range(self.ap.capwap_MaxRetransmit):
            if not need_assoc_resp_clients:
                break
            RetransmitInterval *= 2
            tx_pkts = []
            for client in list(need_assoc_resp_clients.values()):
                tx_pkt = CAPWAP_PKTS.client_assoc(
                    self.ap, slot_id=0, client_mac=client.mac)
                tx_pkts.append(self.ap.wrap_capwap_pkt(tx_pkt, dst_port=5247))
            yield('put', tx_pkts)

            timer = PassiveTimer(RetransmitInterval)
            while not timer.has_expired():
                if not need_assoc_resp_clients:
                    break
                rx_pkt = yield('get', RetransmitInterval)
                if rx_pkt and isinstance(rx_pkt, Dot11_swapped):
                    if self.ap.is_debug:
                        rx_pkt.show2()
                    if rx_pkt.type == 0 and rx_pkt.subtype == 1 and rx_pkt.addr1 in need_assoc_resp_clients:
                        self.log('Received Association response for %s.' %
                                 need_assoc_resp_clients[rx_pkt.addr1].ip)
                        need_assoc_resp_clients[rx_pkt.addr1].is_associated = True
                        del need_assoc_resp_clients[rx_pkt.addr1]

        if need_assoc_resp_clients:
            yield('err', 'No Association response for clients: %s' % ', '.join([client.ip for client in list(need_assoc_resp_clients.values())]))

        need_arp_resp_clients = dict(
            [(client.ip, client) for client in self.clients])

        RetransmitInterval = self.ap.capwap_RetransmitInterval
        for _ in range(self.ap.capwap_MaxRetransmit):
            if not need_arp_resp_clients:
                return

            RetransmitInterval *= 2
            tx_pkts = []
            for client in list(need_arp_resp_clients.values()):
                tx_pkts.append(self.ap.wrap_pkt_by_wlan(
                    client, self.ap.get_garp_pkt(client)))
                tx_pkts.append(self.ap.wrap_pkt_by_wlan(client, self.ap.get_arp_pkt(
                    client, 'who-has', 'ff:ff:ff:ff:ff:ff', self.ap.ip_dst)))
            yield('put', tx_pkts)

            timer = PassiveTimer(RetransmitInterval)
            while not timer.has_expired():
                if not need_arp_resp_clients:
                    return
                rx_pkt = yield('get', RetransmitInterval)
                if rx_pkt and isinstance(rx_pkt, Dot11_swapped) and ARP in rx_pkt:
                    rx_arp = rx_pkt[ARP]
                    client = need_arp_resp_clients.get(rx_arp.pdst)
                    if not client or client.mac != rx_arp.hwdst or client.mac != rx_pkt[Dot11_swapped].addr1:
                        continue
                    #print 'client for arp resp'
                    if rx_arp.hwsrc != self.ap.mac_dst:
                        self.log('MAC address returned from ARP(%s) differs from WLC MAC(%s)' % (
                            rx_arp.hwsrc, self.ap.mac_dst))
                    client.seen_arp_reply = True
                    del need_arp_resp_clients[rx_arp.pdst]
                    self.log('Got WLC response with MAC: %s' % rx_arp.hwsrc)


class STLServiceApShutdownDTLS(STLServiceAp):
    def run(self, pipe):
        for client in self.ap.clients:
            client.disconnect()
        if not self.ap.is_dtls_established():
            return
        with self.ap.ssl_lock:
            libssl.SSL_shutdown(self.ap.ssl)
        tx_pkt = self.ap.wrap_capwap_pkt(b'\1\0\0\0' + self.ap.ssl_read())
        self.ap.reset_vars()
        yield pipe.async_tx_pkt(PacketBuffer(tx_pkt))



#!/usr/bin/python3

"""
ICMP service implementation

Description:
    <FILL ME HERE>

How to use:
    <FILL ME HERE>

Author:
  Itay Marom

"""
import random
from collections import defaultdict

from scapy.layers.inet import ICMP, IP
from scapy.layers.l2 import Ether

from trex_stl_lib.trex_stl_exceptions import STLError

from trex_stl_lib.trex_stl_vlan import VLAN

from .trex_stl_service import STLService, STLServiceFilter


class STLServiceFilterICMP(STLServiceFilter):
    '''
        Service filter for ICMP services
    '''

    def __init__(self):
        self.services = defaultdict(list)

    def add(self, service):
        # an ICMP service can be identified by src IP, identifier and seq
        self.services[(service.src_ip, service.id, service.seq,
                       tuple(service.vlan))].append(service)

    def lookup(self, pkt):
        scapy_pkt = Ether(pkt)

        # not ICMP
        if 'ICMP' not in scapy_pkt:
            return []

        vlans = VLAN.extract(scapy_pkt)

        # ignore VLAN 0
        vlans = vlans if vlans != [0] else []

        src_ip = scapy_pkt['IP'].dst
        id = scapy_pkt['ICMP'].id
        seq = scapy_pkt['ICMP'].seq

        return self.services.get((src_ip, id, seq, tuple(vlans)), [])

    def get_bpf_filter(self):
        # a simple BPF pattern for ICMP(duplicate for QinQ)
        return 'icmp or(vlan and icmp) or(vlan and icmp)'


class STLServiceICMP(STLService):
    '''
        ICMP service - generate echo requests
    '''

    def __init__(self, ctx, dst_ip, src_ip=None, pkt_size=64, timeout_sec=3, verbose_level=STLService.ERROR, vlan=None):

        # init the base object
        super(STLServiceICMP, self).__init__(verbose_level)

        if src_ip is None:
            src_ip = ctx.get_src_ipv4()
            if not src_ip:
                raise STLError('PING: port {} does not have an IPv4 address. please manually provide source IPv4'.format(
                    ctx.get_port_id()))

        self.src_ip = src_ip
        self.dst_ip = dst_ip
        self.vlan = VLAN(vlan)

        self.pkt_size = pkt_size
        self.timeout_sec = timeout_sec

        self.id = random.getrandbits(16)
        self.seq = 0

        self.record = None

    def get_filter_type(self):
        return STLServiceFilterICMP

    def run(self, pipe):
        '''
            Will execute ICMP echo request
        '''

        self.record = None

        self.log(
            "ICMP: {:<15} ---> Pinging '{}'".format(self.src_ip, self.dst_ip))

        base_pkt = Ether()/IP(src=self.src_ip, dst=self.dst_ip)/ICMP(id=self.id, type=8)
        self.vlan.embed(base_pkt)

        pad = max(0, self.pkt_size - len(base_pkt))
        pkt = base_pkt / ('x' * pad)

        # wait until packet was actually sent - to get the ts
        tx_info = yield pipe.async_tx_pkt(pkt)

        # wait for RX packet
        pkts = yield pipe.async_wait_for_pkt(time_sec=self.timeout_sec)
        if not pkts:
            # timeout - create an empty record
            self.record = self.PINGRecord()
            return

        # take the first one
        response = pkts[0]

        # parse record
        self.record = self.PINGRecord(
            Ether(response['pkt']), tx_info['ts'], response['ts'])

        # log and exit
        self.log('ICMP: {:<15} <--- {}'.format(self.src_ip, str(self.record)))
        return

    def get_record(self):
        return self.record

    class PINGRecord(object):
        TIMEOUT, UNREACHABLE, SUCCESS = list(range(3))

        ICMP_TYPE_ECHO_REPLY = 0
        ICMP_TYPE_DEST_UNREACHABLE = 3

        def __init__(self, scapy_pkt=None, tx_ts=None, rx_ts=None):

            # default values
            self.responder_ip = 'N/A'
            self.ttl = 'N/A'
            self.rtt = 'N/A'
            self.pkt_size = 'N/A'
            self.state = self.TIMEOUT

            if not scapy_pkt:
                return

            if scapy_pkt['ICMP'].type == self.ICMP_TYPE_ECHO_REPLY:
                self.responder_ip = scapy_pkt['IP'].src
                self.ttl = scapy_pkt['IP'].ttl
                self.rtt = (rx_ts - tx_ts) * 1000
                self.pkt_size = len(scapy_pkt)
                self.state = self.SUCCESS
                return

            if scapy_pkt['ICMP'].type == self.ICMP_TYPE_DEST_UNREACHABLE:
                self.responder_ip = scapy_pkt['IP'].src
                self.state = self.ICMP_TYPE_DEST_UNREACBLE
                return

        def __str__(self):

            if self.state == self.SUCCESS:
                return 'Reply from {0}: bytes={1}, time={2:.2f}ms, TTL={3}'.format(self.responder_ip, self.pkt_size, self.rtt, self.ttl)

            elif self.state == self.TIMEOUT:
                return 'Request timed out.'

            elif self.state == self.ICMP_TYPE_DEST_UNREACHABLE:
                return 'Reply from {0}: Destination host unreachable'.format(self.responder_ip)

            assert(0)



#!/usr/bin/python3

# client and exceptions
from trex_stl_lib.trex_stl_exceptions import *
from trex_stl_lib.trex_stl_client import STLClient, LoggerApi

# streams
from trex_stl_lib.trex_stl_streams import *

# packet builder
from trex_stl_lib.trex_stl_packet_builder_scapy import *
from scapy.all import *


# simulator
from trex_stl_lib.trex_stl_sim import STLSim

# std lib (various lib functions)
from trex_stl_lib.trex_stl_std import *


#!/usr/bin/python3

import sys

# from text_opts import format_text

from texttable import Texttable


class TRexTextTable(Texttable):

    def __init__(self):
        Texttable.__init__(self)
        # set class attributes so that it'll be more like TRex standard output
        self.set_chars(['-', '|', '-', '-'])
        self.set_deco(Texttable.HEADER | Texttable.VLINES)


class TRexTextInfo(Texttable):

    def __init__(self):
        Texttable.__init__(self)
        # set class attributes so that it'll be more like TRex standard output
        self.set_chars(['-', ':', '-', '-'])
        self.set_deco(Texttable.VLINES)


def generate_trex_stats_table():
    pass


def print_table_with_header(texttable_obj, header="", untouched_header="",
                            buffer=sys.stdout):
    header = header.replace("_", " ").title() + untouched_header
    # print(format_text(header, 'cyan', 'underline') + "\n", file=buffer)
    print(header)

    # print((texttable_obj.draw() + "\n"), file=buffer)
    print((texttable_obj.draw()))


if __name__ == "__main__":
    pass



#!/usr/bin/python3

"""
Base API for services


Description:
  Base classes used to implement a service

Author:
  Itay Marom

"""

#########################
#
# STL service filter
#
#
#########################


class STLServiceFilter(object):
    '''
        Abstract class for service filtering
        each class of services should
        implement a filter
    '''

    def add(self, service):
        '''
            Adds a service to the filter
        '''
        raise NotImplementedError

    def lookup(self, pkt):
        '''
            Given a 'pkt' return a list
            of services that should get this packet

            can be an empty list
        '''
        raise NotImplementedError

    def get_bpf_filter(self):
        '''
            Each filter needs to describe a BPF filter
            any packets matching the BPF pattern will
            be forwarded to the filter
        '''
        raise NotImplementedError


#########################
#
# STL service
#
#
#########################
class STLService(object):
    '''
        Abstract class for implementing a service
    '''

    ERROR = 3
    WARN = 2
    INFO = 1

    def __init__(self, verbose_level=ERROR):

        # by default, set the service verbose level to error
        self.verbose_level = verbose_level


######### implement-needed functions #########


    def get_filter_type(self):
        '''
            Returns a filter class type
            The filter will manage packet
            forwarding for the services
            in this group
        '''
        raise NotImplementedError

    def run(self, pipe):
        '''
            Executes the service in a run until completion
            model
        '''
        raise NotImplementedError


######### API          #########

    def err(self, msg):
        '''
            Genereate an error
        '''
        raise STLError(msg)

    def set_verbose(self, level):
        '''
            Sets verbose level
        '''
        self.verbose_level = level

    def log(self, msg, level=INFO):
        '''
            Log a message if the level
            is high enough
        '''
        if level >= self.verbose_level:
            print(msg)



#!/usr/bin/python3

from scapy.layers.l2 import Ether
from scapy.layers.inet import IP, UDP
from scapy.packet import NoPayload
from collections import namedtuple

import struct

# a quick parser which uses scapy to generate an offline offset
# tables.
# this way, we get flexibility like scapy but much faster speed
#
# parsers can derive from this parser and enhance it


class ParserError(Exception):
    '''
        throwed when an error happens in the parser
    '''
    pass


class FastParser(object):

    '''
        A fast parser based on scapy
        it gets a base packet as a template
        and a list of fields to be used for parsing

        any packet being parsed by this parser should
        match the template packet
    '''

    def __init__(self, base_pkt):
        self.base_pkt = base_pkt

        # build the base packet to generate fields offsets
        self.base_pkt.build()

        self.pkt_bytes = bytes(base_pkt)

        # by default, no fields are monitored
        self.fields = {}

    def add_field(self, fullname, name, offset=None, sz=None, fmt=None, getter=None, setter=None):
        '''
            adds a new field to the parser
            this field will be accessible when doing 'parse'
        '''

        info = self.__get_field_info(fullname)

        info['name'] = name
        info['fullname'] = fullname

        # override offset if specified by user
        if offset is not None:
            info['offset'] = offset

        # override size if specified by user
        if sz is not None:
            info['sz'] = sz

        # override format if specified by user
        if fmt is not None:
            info['fmt'] = fmt

        elif info['fmt'] == '!4s':
            info['fmt'] = '!I'

        elif info['fmt'] == '!2s':
            info['fmt'] = '!H'

        # use custom functions if those were defined
        info['getter'] = ParserInstance.def_getter if not getter else getter
        info['setter'] = ParserInstance.def_setter if not setter else setter

        # add the field
        self.fields[name] = info

    def parse(self, pkt_bytes):
        '''
            Parse a packet based on the template
            returns a parser instace object with all the monitored fields
        '''
        return ParserInstance(pkt_bytes, self.fields)

    def clone(self):
        '''
            Clones the base packet(template)
            used for manipulating the base packet for packet generation
        '''
        return ParserInstance(self.pkt_bytes, self.fields)

    def __get_field_info(self, field):
        '''
            Internal function
            used to generate all the data per field
        '''

        p = self.base_pkt
        while p is not None and not isinstance(p, NoPayload):
            for f in p.fields_desc:
                if field == "{}.{}".format(p.name, f.name):
                    return {'offset': p._offset + f._offset, 'size': f.sz, 'fmt': f.fmt}

            p = p.payload

        raise ValueError('unknown field: {0}'.format(field))


class ParserInstance(object):
    '''
        Parser instance.
        generated when a packet is parsed or cloned.

        Contains all the monitored fields as attributes which can be read/write
    '''

    def __init__(self, pkt_bytes, fields):
        self.__dict__['pkt_bytes'] = pkt_bytes
        self.__dict__['fields'] = dict(fields)
        self.__dict__['cache'] = {}

    def __getattr__(self, name):

        if not name in self.fields:
            raise ValueError(
                "field '{0}' is not registered under the parser".format(name))

        # multiple gets will hit the cache - no need to parse again
        if name in self.cache:
            return self.cache[name]

        # get the record and fetch the value, then save it to the cache
        info = self.fields[name]
        value = info['getter'](self.pkt_bytes, info)
        self.cache[name] = value

        return value

    def __setattr__(self, name, value):

        if not name in self.fields:
            raise ValueError(
                'field {0} is not registered under the parser'.format(name))

        # invalidate from the cache as we are writing
        if name in self.cache:
            del self.cache[name]

        info = self.fields[name]
        self.__dict__['pkt_bytes'] = info['setter'](
            self.pkt_bytes, info, value)

    def raw(self):
        return self.pkt_bytes

    def show2(self):
        Ether(self.pkt_bytes).show2()

    def fix_chksum(self):
        if not 'ihl' in self.fields:
            raise ParserError(
                "'ihl' field must be registered under the parser for checksum fix")

        ihl = self.ihl & 0xf
        ip_start = self.fields['ihl']['offset']

        # clear the current checksum
        self.chksum = 0

        # init accumulator
        sum = 0

        for i in range(0, ihl * 4, 2):
            word = struct.unpack_from(
                "!H", self.pkt_bytes, offset=ip_start + i)[0]
            sum += word

        self.chksum = (~((sum & 0xffff) + (sum >> 16))) & 0xffff

    @staticmethod
    def def_getter(pkt_bytes, info):
        '''
            Default field getter
            returns None if the offset is outside the boundary
        '''

        min_size = info['offset'] + struct.calcsize(info['fmt'])

        if len(pkt_bytes) < min_size:
            return None

        return struct.unpack_from(info['fmt'], pkt_bytes, info['offset'])[0]

    @staticmethod
    def def_setter(pkt_bytes, info, value):
        '''
            Default field setter
        '''

        # sanity
        min_size = info['offset'] + struct.calcsize(info['fmt'])

        if len(pkt_bytes) < min_size:
            raise ParserError("packet length is '{0}' but setting '{1}' requires at least '{2}' bytes".format(
                len(pkt_bytes), info['name'], min_size))

        packed = struct.pack(info['fmt'], value)
        return pkt_bytes[:info['offset']] + packed + pkt_bytes[info['offset'] + len(packed):]



#!/usr/bin/python3

"""
Handles VLANs

Author:
  Itay Marom

"""
from scapy.layers.l2 import Dot1AD, Dot1Q

from trex_stl_lib.trex_stl_exceptions import STLError


class VLAN(object):
    '''
        A simple class to handle VLANs
    '''

    def __init__(self, vlan):
        # move to canonical form

        if vlan is None:
            # no VLAN and mark it as a default value
            self.is_def = True
            self.tags = ()
            return

        if isinstance(vlan, VLAN):
            # copy constructor
            self.tags = vlan.tags
            self.is_def = vlan.is_def
            return

        # make list of integer
        vlan_list = (vlan,) if isinstance(vlan, int) else vlan

        if len(vlan_list) > 2:
            raise STLError("only up to two VLAN tags are supported")

        for tag in vlan_list:
            if not type(tag) == int:
                raise STLError(
                    "invalid VLAN tag: '{0}'(int value expected)".format(tag))

            if not(tag in range(1, 4096)):
                raise STLError(
                    "invalid VLAN tag: '{0}'(valid range: 1 - 4095)".format(tag))

        self.tags = tuple(vlan_list)
        self.is_def = False

    def __bool__(self):
        return len(self.tags) > 0

    # def __bool__(self):
    #     return self.__nonzero__()

    def __iter__(self):
        return self.tags.__iter__()

    def get_tags(self):
        return self.tags

    def is_default(self):
        '''
            returns True if no values were provided during
            the object creation
            it represents an empty VLAN as a default value
        '''
        return self.is_def

    def get_desc(self):

        if len(self.tags) == 1:
            return "VLAN '{0}'".format(self.tags[0])
        elif len(self.tags) == 2:
            return "QinQ '{0}/{1}'".format(self.tags[0], self.tags[1])
        else:
            return ''

    @staticmethod
    def extract(scapy_pkt):
        '''
            Given a scapy packet, returns all the VLANs
            in the packet
        '''

        # no VLANs
        if scapy_pkt.type not in(0x8100, 0x88a8):
            return []

        vlans = []
        vlan_layer = scapy_pkt.payload

        while type(vlan_layer) in(Dot1Q, Dot1AD):
            # append
            vlans.append(vlan_layer.vlan)

            # next
            vlan_layer = vlan_layer.payload

        return vlans

    def embed(self, scapy_pkt):
        '''
            Given a scapy packet, embedd the VLAN config
            into the packet
        '''
        if not self.tags:
            return

        ether = scapy_pkt.getlayer(0)
        ether_payload = scapy_pkt.payload

        # single VLAN
        if len(self.tags) == 1:
            vlan = Dot1Q(vlan=self.tags[0])
            vlan.payload = ether.payload
            ether.payload = vlan

        # dobule VLAN
        elif len(self.tags) == 2:
            dot1ad = Dot1AD(vlan=self.tags[0])
            dot1q = Dot1Q(vlan=self.tags[1])
            dot1ad.payload = dot1q
            dot1q.payload = ether_payload

            ether.payload = dot1ad



#!/usr/bin/python3

import argparse
import inspect
import os
import re
from collections import namedtuple

from trex_stl_lib.trex_stl_types import RC_ERR, STLError, listify

from trex_stl_lib.trex_stl_vlan import VLAN

from .common import (is_valid_ipv4, is_valid_ipv6, is_valid_mac,
                     list_difference, list_remove_dup)
from .constants import FLOW_CTRL_DICT, ON_OFF_DICT, UP_DOWN_DICT
from .text_opts import bold, format_text

ArgumentPack = namedtuple('ArgumentPack', ['name_or_flags', 'options'])
ArgumentGroup = namedtuple('ArgumentGroup', ['type', 'args', 'options'])


# list of available parsing options
# _constants = '''
'''

MULTIPLIER
MULTIPLIER_STRICT
PORT_LIST
ALL_PORTS
PORT_LIST_WITH_ALL
FILE_PATH
FILE_FROM_DB
SERVER_IP
STREAM_FROM_PATH_OR_FILE
DURATION
TIMEOUT
FORCE
READONLY
DRY_RUN
SYNCHRONIZED
XTERM
TOTAL
FULL_OUTPUT
IPG
MIN_IPG
SPEEDUP
COUNT
PROMISCUOUS
MULTICAST
LINK_STATUS
LED_STATUS
TUNABLES
REMOTE_FILE
LOCKED
PIN_CORES
CORE_MASK
DUAL
FLOW_CTRL
SUPPORTED
FILE_PATH_NO_CHECK

OUTPUT_FILENAME
LIMIT
PORT_RESTART

RETRIES

SINGLE_PORT
DST_MAC

PING_IP
PING_COUNT
PKT_SIZE

SERVICE_OFF

TX_PORT_LIST
RX_PORT_LIST

SRC_IPV4
DST_IPV4

CAPTURE_ID

SCAPY_PKT
SHOW_LAYERS
SCAPY_PKT_CMD

GLOBAL_STATS
PORT_STATS
PORT_STATUS
STREAMS_STATS
STATS_MASK
CPU_STATS
MBUF_STATS
EXTENDED_STATS
EXTENDED_INC_ZERO_STATS

STREAMS_MASK
CORE_MASK_GROUP
CAPTURE_PORTS_GROUP

MONITOR_TYPE_VERBOSE
MONITOR_TYPE_PIPE
MONITOR_TYPE
BPF_FILTER

VLAN_TAGS
CLEAR_VLAN
VLAN_CFG

PLUGIN_NAME

# ALL_STREAMS
# STREAM_LIST_WITH_ALL

# list of ArgumentGroup types
MUTEX
NON_MUTEX

'''
# Wacky constant assigning code, replacing with enum
#
# for index, line in enumerate(_constants.splitlines()):
#     var = line.strip().split()
#    if not var or '#' in var[0]:
#        continue
#    exec('%s = %s' % (var[0], index))

MULTIPLIER = 2
MULTIPLIER_STRICT = 3
PORT_LIST = 4
ALL_PORTS = 5
PORT_LIST_WITH_ALL = 6
FILE_PATH = 7
FILE_FROM_DB = 8
SERVER_IP = 9
STREAM_FROM_PATH_OR_FILE = 10
DURATION = 11
TIMEOUT = 12
FORCE = 13
READONLY = 14
DRY_RUN = 15
SYNCHRONIZED = 16
XTERM = 17
TOTAL = 18
FULL_OUTPUT = 19
IPG = 20
MIN_IPG = 21
SPEEDUP = 22
COUNT = 23
PROMISCUOUS = 24
MULTICAST = 25
LINK_STATUS = 26
LED_STATUS = 27
TUNABLES = 28
REMOTE_FILE = 29
LOCKED = 30
PIN_CORES = 31
CORE_MASK = 32
DUAL = 33
FLOW_CTRL = 34
SUPPORTED = 35
FILE_PATH_NO_CHECK = 36
OUTPUT_FILENAME = 38
LIMIT = 39
PORT_RESTART = 40
RETRIES = 42
SINGLE_PORT = 44
DST_MAC = 45
PING_IP = 47
PING_COUNT = 48
PKT_SIZE = 49
SERVICE_OFF = 51
TX_PORT_LIST = 53
RX_PORT_LIST = 54
SRC_IPV4 = 56
DST_IPV4 = 57
CAPTURE_ID = 59
SCAPY_PKT = 61
SHOW_LAYERS = 62
SCAPY_PKT_CMD = 63
GLOBAL_STATS = 65
PORT_STATS = 66
PORT_STATUS = 67
STREAMS_STATS = 68
STATS_MASK = 69
CPU_STATS = 70
MBUF_STATS = 71
EXTENDED_STATS = 72
EXTENDED_INC_ZERO_STATS = 73
STREAMS_MASK = 75
CORE_MASK_GROUP = 76
CAPTURE_PORTS_GROUP = 77
MONITOR_TYPE_VERBOSE = 79
MONITOR_TYPE_PIPE = 80
MONITOR_TYPE = 81
BPF_FILTER = 82
VLAN_TAGS = 84
CLEAR_VLAN = 85
VLAN_CFG = 86
PLUGIN_NAME = 88
MUTEX = 94
NON_MUTEX = 95


def check_negative(value):
    ivalue = int(value)
    if ivalue < 0:
        raise argparse.ArgumentTypeError(
            "non positive value provided: '{0}'".format(value))
    return ivalue


def match_time_unit(val):
    '''match some val against time shortcut inputs '''
    match = re.match(r"^(\d+(\.\d+)?)([m|h]?)$", val)
    if match:
        digit = float(match.group(1))
        unit = match.group(3)
        if not unit:
            return digit
        elif unit == 'm':
            return digit*60
        else:
            return digit*60*60
    else:
        raise argparse.ArgumentTypeError("Duration should be passed in the following format: \n"
                                         "-d 100 : in sec \n"
                                         "-d 10m : in min \n"
                                         "-d 1h  : in hours")


match_multiplier_help = """Multiplier should be passed in the following format:
                          [number][<empty> | bps | kbps | mbps |  gbps | pps | kpps | mpps | %% ].

                          no suffix will provide an absoulute
                          factor and percentage
                          will provide a percentage of the line rate. examples

                          '-m 10',
                          '-m 10kbps',
                          '-m 10kbpsl1',
                          '-m 10mpps',
                          '-m 23%% '

                          '-m 23%%' : is 23%% L1 bandwidth
                          '-m 23mbps': is 23mbps in L2 bandwidth
                                       (including FCS+4)
                          '-m 23mbpsl1': is 23mbps in L1 bandwidth

                          """


# decodes multiplier
# if allow_update - no +/- is allowed
# divide states between how many entities the
# value should be divided
def decode_multiplier(val, allow_update=False, divide_count=1):

    factor_table = {None: 1, 'k': 1e3, 'm': 1e6, 'g': 1e9}
    pattern = r"^(\d+(\.\d+)?)(((k|m|g)?(bpsl1|pps|bps))|%)?"

    # do we allow updates ?  +/-
    if not allow_update:
        pattern += "$"
        match = re.match(pattern, val)
        op = None
    else:
        pattern += r"([\+\-])?$"
        match = re.match(pattern, val)
        if match:
            op = match.group(7)
        else:
            op = None

    result = {}

    if not match:
        return None

    # value in group 1
    value = float(match.group(1))

    # decode unit as whole
    unit = match.group(3)

    # k,m,g
    factor = match.group(5)

    # type of multiplier
    m_type = match.group(6)

    # raw type(factor)
    if not unit:
        result['type'] = 'raw'
        result['value'] = value

    # percentage
    elif unit == '%':
        result['type'] = 'percentage'
        result['value'] = value

    elif m_type == 'bps':
        result['type'] = 'bps'
        result['value'] = value * factor_table[factor]

    elif m_type == 'pps':
        result['type'] = 'pps'
        result['value'] = value * factor_table[factor]

    elif m_type == 'bpsl1':
        result['type'] = 'bpsl1'
        result['value'] = value * factor_table[factor]

    if op == "+":
        result['op'] = "add"
    elif op == "-":
        result['op'] = "sub"
    else:
        result['op'] = "abs"

    if result['op'] != 'percentage':
        result['value'] = result['value'] / divide_count

    return result


def match_multiplier(val):
    '''match some val against multiplier  shortcut inputs '''
    result = decode_multiplier(val, allow_update=True)
    if not result:
        raise argparse.ArgumentTypeError(match_multiplier_help)

    return val


def match_multiplier_strict(val):
    '''match some val against multiplier  shortcut inputs '''
    result = decode_multiplier(val, allow_update=False)
    if not result:
        raise argparse.ArgumentTypeError(match_multiplier_help)

    return val


def hex_int(val):
    pattern = r"0x[1-9a-fA-F][0-9a-fA-F]*"

    if not re.match(pattern, val):
        raise argparse.ArgumentTypeError(
            "{0} is not a valid positive HEX formatted number".format(val))

    return int(val, 16)


def action_check_vlan():
    class VLANCheck(argparse.Action):
        def __call__(self, parser, args, values, option_string=None):
            try:
                vlan = VLAN(values)
            except STLError as e:
                parser.error(e.brief())

            setattr(args, self.dest, vlan.get_tags())

            return

    return VLANCheck


def action_bpf_filter_merge():
    class BPFFilterMerge(argparse.Action):
        def __call__(self, parser, args, values, option_string=None):
            setattr(args, self.dest, ' '.join(values).strip("'\""))
            return

    return BPFFilterMerge


def is_valid_file(filename):
    if os.path.isdir(filename):
        raise argparse.ArgumentTypeError(
            "Given path '%s' is a directory" % filename)
    if not os.path.isfile(filename):
        raise argparse.ArgumentTypeError(
            "The file '%s' does not exist" % filename)

    return filename

# scapy decoder class for parsing opts


class ScapyDecoder(object):
    scapy_layers = None

    @staticmethod
    def init():
        # one time
        if ScapyDecoder.scapy_layers:
            return

        import scapy.all
        import scapy.layers.dhcp

        raw = {}

        # default layers
        raw.update(scapy.all.__dict__)

        # extended layers - add here
        raw.update(scapy.layers.dhcp.__dict__)

        ScapyDecoder.scapy_layers = {k: v for k, v in list(
            raw.items()) if (inspect.isclass(v)
                             and issubclass(v, scapy.all.Packet))}

    @staticmethod
    def to_scapy(scapy_str):
        ScapyDecoder.init()

        try:
            scapy_obj = eval(scapy_str, {
                             '__builtins__': {}, 'True': True, 'False': False},
                             ScapyDecoder.scapy_layers)
            len(scapy_obj)
            return scapy_obj
        except Exception as e:
            raise argparse.ArgumentTypeError(
                "invalid scapy expression: '{0}' - {1}"
                .format(scapy_str, str(e)))

    @staticmethod
    def formatted_layers():
        ScapyDecoder.init()

        output = ''
        for k, v in sorted(ScapyDecoder.scapy_layers.items()):
            name = format_text("'{}'".format(k), 'bold')
            descr = v.name
            # fields  = ", ".join(["'%s'" % f.name for f in v.fields_desc])
            # print("{:<50} - {}({})".format(name ,descr, fields))
            output += "{:<50} - {}\n".format(name, descr)

        return output


def check_ipv4_addr(ipv4_str):
    if not is_valid_ipv4(ipv4_str):
        raise argparse.ArgumentTypeError(
            "invalid IPv4 address: '{0}'".format(ipv4_str))

    return ipv4_str


def check_ip_addr(addr):
    if not(is_valid_ipv4(addr) or is_valid_ipv6(addr)):
        raise argparse.ArgumentTypeError(
            "invalid IPv4/6 address: '{0}'".format(addr))

    return addr


def check_pkt_size(pkt_size):
    try:
        pkt_size = int(pkt_size)
    except ValueError:
        raise argparse.ArgumentTypeError(
            "invalid packet size type: '{0}'".format(pkt_size))

    if(pkt_size < 64) or(pkt_size > 9216):
        raise argparse.ArgumentTypeError(
            "invalid packet size: '{0}' - valid range is 64 to 9216"
            .format(pkt_size))

    return pkt_size


def check_mac_addr(addr):
    if not is_valid_mac(addr):
        raise argparse.ArgumentTypeError(
            "not a valid MAC address: '{0}'".format(addr))

    return addr


def decode_tunables(tunable_str):
    tunables = {}

    # split by comma to tokens
    tokens = tunable_str.split(',')

    # each token is of form X=Y
    for token in tokens:
        m = re.search(r'(\S+)=(.+)', token)
        if not m:
            raise argparse.ArgumentTypeError(
                "bad syntax for tunables: {0}".format(token))
        val = m.group(2)           # string
        # need to remove the quotes from value
        if val.startswith(("'", '"')) and val.endswith(("'", '"')) and len(val) > 1:
            val = val[1:-1]
        elif val.startswith('0x'):  # hex
            val = int(val, 16)
        else:
            try:
                if '.' in val:     # float
                    val = float(val)
                else:              # int
                    val = int(val)
            except Exception:
                pass
        tunables[m.group(1)] = val

    return tunables


OPTIONS_DB = {MULTIPLIER: ArgumentPack(['-m', '--multiplier'],
                                       {'help': match_multiplier_help,
                                        'dest': "mult",
                                                'default': "1",
                                                'type': match_multiplier}),

              MULTIPLIER_STRICT: ArgumentPack(['-m', '--multiplier'],
                                              {'help': match_multiplier_help,
                                               'dest': "mult",
                                                       'default': "1",
                                               'type': match_multiplier_strict}),

              TOTAL: ArgumentPack(['-t', '--total'],
                                  {'help': "traffic will be divided "
                                           + "between all ports specified",
                                           'dest': "total",
                                           'default': False,
                                           'action': "store_true"}),

              IPG: ArgumentPack(['-i', '--ipg'],
                                {'help': "IPG value in usec between packets. "
                                         + "default will be from the pcap",
                                         'dest': "ipg_usec",
                                         'default':  None,
                                         'type': float}),

              MIN_IPG: ArgumentPack(['--min-ipg'],
                                    {'help': "Minimal IPG value in usec "
                                             + "between packets. Used to "
                                             + "guard from too small IPGs.",
                                             'dest': "min_ipg_usec",
                                             'default':  None,
                                             'type': float}),

              SPEEDUP: ArgumentPack(['-s', '--speedup'],
                                    {'help': "Factor to accelerate the injection. "
                                             + "effectively means IPG = IPG / SPEEDUP",
                                             'dest': "speedup",
                                             'default':  1.0,
                                             'type': float}),

              COUNT: ArgumentPack(['-c', '--count'],
                                  {'help': "How many times to perform action "
                                           + "[default is 1, 0 means forever]",
                                           'dest': "count",
                                           'default':  1,
                                           'type': int}),

              PROMISCUOUS: ArgumentPack(['--prom'],
                                        {'help': "Set port promiscuous on/off",
                                                 'choices': ON_OFF_DICT}),

              MULTICAST: ArgumentPack(['--mult'],
                                      {'help': "Set port multicast on/off",
                                               'choices': ON_OFF_DICT}),

              LINK_STATUS: ArgumentPack(['--link'],
                                        {'help': 'Set link status up/down',
                                                 'choices': UP_DOWN_DICT}),

              LED_STATUS: ArgumentPack(['--led'],
                                       {'help': 'Set LED status on/off',
                                                'choices': ON_OFF_DICT}),

              FLOW_CTRL: ArgumentPack(['--fc'],
                                      {'help': 'Set Flow Control type',
                                               'dest': 'flow_ctrl',
                                               'choices': FLOW_CTRL_DICT}),

              SRC_IPV4: ArgumentPack(['--src'],
                                     {'help': 'Configure source IPv4 address',
                                              'dest': 'src_ipv4',
                                              'required': True,
                                              'type': check_ipv4_addr}),

              DST_IPV4: ArgumentPack(['--dst'],
                                     {'help': 'Configure destination IPv4 address',
                                              'dest': 'dst_ipv4',
                                              'required': True,
                                              'type': check_ipv4_addr}),


              DST_MAC: ArgumentPack(['--dst'],
                                    {'help': 'Configure destination MAC address',
                                             'dest': 'dst_mac',
                                             'required': True,
                                             'type': check_mac_addr}),

              RETRIES: ArgumentPack(['-r', '--retries'],
                                    {'help': 'retries count [default is zero]',
                                             'dest': 'retries',
                                             'default':  0,
                                             'type': int}),


              OUTPUT_FILENAME: ArgumentPack(['-o', '--output'],
                                            {'help': 'Output PCAP filename',
                                                     'dest': 'output_filename',
                                                     'default': None,
                                                     'type': str}),


              PORT_RESTART: ArgumentPack(['-r', '--restart'],
                                         {'help': 'hard restart port(s)',
                                                  'dest': 'restart',
                                                  'default': False,
                                                  'action': 'store_true'}),

              LIMIT: ArgumentPack(['-l', '--limit'],
                                  {'help': 'Limit the packet count to '
                                           + 'be written to the file',
                                           'dest': 'limit',
                                           'default':  1000,
                                           'type': int}),


              SUPPORTED: ArgumentPack(['--supp'],
                                      {'help': 'Show which attributes are '
                                               + 'supported by current NICs',
                                               'default': None,
                                               'action': 'store_true'}),

              TUNABLES: ArgumentPack(['-t'],
                                     {'help': "Sets tunables for a profile. "
                                              + "Example: '-t fsize=100,pg_id=7'",
                                              'metavar': 'T1=VAL[,T2=VAL ...]',
                                      'dest': "tunables",
                                      'default': None,
                                      'action': 'merge',
                                      'type': decode_tunables}),

              PORT_LIST: ArgumentPack(['-p', '--port'],
                                      {"nargs": '+',
                                       'dest': 'ports',
                                               'metavar': 'PORTS',
                                       'action': 'merge',
                                       'type': int,
                                       'help': "A list of ports on which "
                                                          + "to apply the command",
                                       'default': []}),


              SINGLE_PORT: ArgumentPack(['-p', '--port'],
                                        {'dest': 'ports',
                                                 'type': int,
                                                 'metavar': 'PORT',
                                         'help': 'source port for the action',
                                         'required': True}),

              PING_IP: ArgumentPack(['-d'],
                                    {'help': 'which IPv4/6 to ping',
                                             'dest': 'ping_ip',
                                             'required': True,
                                             'type': check_ip_addr}),

              PING_COUNT: ArgumentPack(['-n', '--count'],
                                       {'help': 'How many times to ping '
                                                + '[default is 5]',
                                                'dest': 'count',
                                                'default':  5,
                                                'type': int}),

              PKT_SIZE: ArgumentPack(['-s'],
                                     {'dest': 'pkt_size',
                                              'help': 'packet size to use',
                                              'default': 64,
                                              'type': check_pkt_size}),


              ALL_PORTS: ArgumentPack(['-a'],
                                      {"action": "store_true",
                                       "dest": "all_ports",
                                               'help': "Set this flag to apply the "
                                               + "command on all available ports",
                                               'default': False},),

              DURATION: ArgumentPack(['-d'],
                                     {'action': "store",
                                      'metavar': 'TIME',
                                      'dest': 'duration',
                                                 'type': match_time_unit,
                                                 'default': -1.0,
                                                 'help': "Set duration time for job."}),

              TIMEOUT: ArgumentPack(['-t'],
                                    {'action': "store",
                                     'metavar': 'TIMEOUT',
                                     'dest': 'timeout',
                                                'type': int,
                                                'default': None,
                                                'help': "Timeout for operation in seconds."}),

              FORCE: ArgumentPack(['--force'],
                                  {"action": "store_true",
                                   'default': False,
                                   'help': "Set if you want to stop active "
                                           + "ports before appyling command."}),

              READONLY: ArgumentPack(['-r'],
                                     {'action': 'store_true',
                                      'dest': 'readonly',
                                              'help': 'Do not acquire ports, '
                                              + 'connect as read-only.'}),

              REMOTE_FILE: ArgumentPack(['-r', '--remote'],
                                        {"action": "store_true",
                                         'default': False,
                                         'help': "file path should be interpeted "
                                                 + "by the server(remote file)"}),

              DUAL: ArgumentPack(['--dual'],
                                 {"action": "store_true",
                                  'default': False,
                                  'help': "Transmit in a dual mode - requires "
                                          + "ownership on the adjacent port"}),

              FILE_PATH: ArgumentPack(['-f'],
                                      {'metavar': 'FILE',
                                                  'dest': 'file',
                                                  'nargs': 1,
                                                  'required': True,
                                                  'type': is_valid_file,
                                                  'help': "File path to load"}),

              FILE_PATH_NO_CHECK: ArgumentPack(['-f'],
                                               {'metavar': 'FILE',
                                                           'dest': 'file',
                                                           'nargs': 1,
                                                           'required': True,
                                                           'type': str,
                                                           'help': "File path to load"}),

              FILE_FROM_DB: ArgumentPack(['--db'],
                                         {'metavar': 'LOADED_STREAM_PACK',
                                                     'help': "A stream pack which already "
                                                             + "loaded into console cache."}),

              SERVER_IP: ArgumentPack(['--server'],
                                      {'metavar': 'SERVER',
                                                  'help': "server IP"}),

              DRY_RUN: ArgumentPack(['-n', '--dry'],
                                    {'action': 'store_true',
                                     'dest': 'dry',
                                             'default': False,
                                             'help': "Dry run - no traffic will be injected"}),

              SYNCHRONIZED: ArgumentPack(['--sync'],
                                         {'action': 'store_true',
                                          'dest': 'sync',
                                                  'default': False,
                                                  'help': 'Run the traffic with syncronized '
                                                  + 'time at adjacent ports. '
                                                  + 'Need to ensure effective ipg is '
                                                  + 'at least 1000 usec.'}),

              XTERM: ArgumentPack(['-x', '--xterm'],
                                  {'action': 'store_true',
                                   'dest': 'xterm',
                                           'default': False,
                                           'help': "Starts TUI in xterm window"}),

              LOCKED: ArgumentPack(['-l', '--locked'],
                                   {'action': 'store_true',
                                    'dest': 'locked',
                                            'default': False,
                                            'help': "Locks TUI on legend mode"}),

              FULL_OUTPUT: ArgumentPack(['--full'],
                                        {'action': 'store_true',
                                         'help': "Prompt full info in a JSON format"}),

              GLOBAL_STATS: ArgumentPack(['-g'],
                                         {'action': 'store_true',
                                          'help': "Fetch only global statistics"}),

              PORT_STATS: ArgumentPack(['-p'],
                                       {'action': 'store_true',
                                        'help': "Fetch only port statistics"}),

              PORT_STATUS: ArgumentPack(['--ps'],
                                        {'action': 'store_true',
                                         'help': "Fetch only port status data"}),

              STREAMS_STATS: ArgumentPack(['-s'],
                                          {'action': 'store_true',
                                           'help': "Fetch only streams stats"}),

              CPU_STATS: ArgumentPack(['-c'],
                                      {'action': 'store_true',
                                       'help': "Fetch only CPU utilization stats"}),

              MBUF_STATS: ArgumentPack(['-m'],
                                       {'action': 'store_true',
                                        'help': "Fetch only MBUF utilization stats"}),

              EXTENDED_STATS: ArgumentPack(['-x'],
                                           {'action': 'store_true',
                                            'help': "Fetch xstats of port, "
                                                    + "excluding lines with "
                                                    + "zero values"}),

              EXTENDED_INC_ZERO_STATS: ArgumentPack(['--xz'],
                                                    {'action': 'store_true',
                                                     'help': "Fetch xstats of port, "
                                                             + "including lines with "
                                                             + "zero values"}),

              STREAMS_MASK: ArgumentPack(['--streams'],
                                         {"nargs": '+',
                                          'dest': 'streams',
                                                  'metavar': 'STREAMS',
                                          'type': int,
                                          'help': "A list of stream IDs to query "
                                                             + "about. Default: "
                                                             + "analyze all streams",
                                          'default': []}),


              PIN_CORES: ArgumentPack(['--pin'],
                                      {'action': 'store_true',
                                       'dest': 'pin_cores',
                                               'default': False,
                                               'help': "Pin cores to interfaces - "
                                               + "cores will be divided "
                                               + "between interfaces"
                                               + "(performance boot for "
                                               + "symetric profiles)"}),

              CORE_MASK: ArgumentPack(['--core_mask'],
                                      {'action': 'store',
                                       'nargs': '+',
                                       'type': hex_int,
                                       'dest': 'core_mask',
                                               'default': None,
                                               'help': "Core mask - only cores responding "
                                               + "to the bit mask will be active"}),

              SERVICE_OFF: ArgumentPack(['--off'],
                                        {'action': 'store_false',
                                         'dest': 'enabled',
                                                 'default': True,
                                                 'help': 'Deactivates services on port(s)'}),

              TX_PORT_LIST: ArgumentPack(['--tx'],
                                         {'nargs': '+',
                                          'dest': 'tx_port_list',
                                                  'metavar': 'TX',
                                          'action': 'merge',
                                          'type': int,
                                          'help': 'A list of ports to capture '
                                                             + 'on the TX side',
                                          'default': []}),


              RX_PORT_LIST: ArgumentPack(['--rx'],
                                         {'nargs': '+',
                                          'dest': 'rx_port_list',
                                                  'metavar': 'RX',
                                          'action': 'merge',
                                          'type': int,
                                          'help': 'A list of ports to capture '
                                                             + 'on the RX side',
                                          'default': []}),


              MONITOR_TYPE_VERBOSE: ArgumentPack(['-v', '--verbose'],
                                                 {'action': 'store_true',
                                                  'dest': 'verbose',
                                                          'default': False,
                                                          'help': 'output to screen as '
                                                          + 'verbose'}),

              MONITOR_TYPE_PIPE: ArgumentPack(['-p', '--pipe'],
                                              {'action': 'store_true',
                                               'dest': 'pipe',
                                                       'default': False,
                                                       'help': 'forward packets to a pipe'}),


              BPF_FILTER: ArgumentPack(['-f', '--filter'],
                                       {'type': str,
                                        'nargs': '+',
                                        'action': action_bpf_filter_merge(),
                                        'dest': 'filter',
                                                'default': '',
                                        'help': 'BPF filter'}),


              CAPTURE_ID: ArgumentPack(['-i', '--id'],
                                       {'help': "capture ID to remove",
                                        'dest': "capture_id",
                                        'type': int,
                                        'required': True}),


              SCAPY_PKT: ArgumentPack(['-s'],
                                      {'dest': 'scapy_pkt',
                                               'metavar': 'PACKET',
                                       'type': ScapyDecoder.to_scapy,
                                       'help': 'A scapy notation packet(e.g.: '
                                                          + 'Ether()/IP())'}),

              SHOW_LAYERS: ArgumentPack(['--layers', '-l'],
                                        {'action': 'store_true',
                                         'dest': 'layers',
                                                 'help': "Show all registered layers / "
                                                 + "inspect a specific layer"}),


              VLAN_TAGS: ArgumentPack(['--vlan', '-v'],
                                      {'dest': 'vlan',
                                               'action': action_check_vlan(),
                                               'type': int,
                                               'nargs': '*',
                                               'metavar': 'VLAN',
                                       'help': 'single or double VLAN tags'}),

              CLEAR_VLAN: ArgumentPack(['-c'],
                                       {'action': 'store_true',
                                        'dest': 'clear_vlan',
                                                'default': False,
                                                'help': "clear any VLAN configuration"}),

              PLUGIN_NAME: ArgumentPack(['plugin_name'],
                                        {'type': str,
                                         'metavar': 'name',
                                         'help': 'Name of plugin'}),

              SCAPY_PKT_CMD: ArgumentGroup(MUTEX, [SCAPY_PKT,
                                                   SHOW_LAYERS],
                                           {'required': True}),

              # advanced options
              PORT_LIST_WITH_ALL: ArgumentGroup(MUTEX, [PORT_LIST,
                                                        ALL_PORTS],
                                                {'required': False}),


              VLAN_CFG: ArgumentGroup(MUTEX, [VLAN_TAGS,
                                              CLEAR_VLAN],
                                      {'required': True}),


              STREAM_FROM_PATH_OR_FILE: ArgumentGroup(MUTEX, [FILE_PATH,
                                                              FILE_FROM_DB],
                                                      {'required': True}),
              STATS_MASK: ArgumentGroup(MUTEX, [GLOBAL_STATS,
                                                PORT_STATS,
                                                PORT_STATUS,
                                                STREAMS_STATS,
                                                CPU_STATS,
                                                MBUF_STATS,
                                                EXTENDED_STATS,
                                                EXTENDED_INC_ZERO_STATS],
                                        {}),


              CORE_MASK_GROUP:  ArgumentGroup(MUTEX, [PIN_CORES,
                                                      CORE_MASK],
                                              {'required': False}),

              CAPTURE_PORTS_GROUP: ArgumentGroup(NON_MUTEX, [TX_PORT_LIST,
                                                             RX_PORT_LIST],
                                                 {}),


              MONITOR_TYPE: ArgumentGroup(MUTEX, [MONITOR_TYPE_VERBOSE,
                                                  MONITOR_TYPE_PIPE],
                                          {'required': False}),

              }


class _MergeAction(argparse._AppendAction):
    def __call__(self, parser, namespace, values, option_string=None):
        items = getattr(namespace, self.dest)
        if not items:
            items = values
        elif type(items) is list and type(values) is list:
            items.extend(values)
        elif type(items) is dict and type(values) is dict:  # tunables are dict
            items.update(values)
        else:
            raise Exception(
                "Argparser 'merge' option should be used on dict or list.")

        setattr(namespace, self.dest, items)


class CCmdArgParser(argparse.ArgumentParser):

    def __init__(self, stateless_client=None, *args, **kwargs):
        super(CCmdArgParser, self).__init__(*args, **kwargs)
        self.stateless_client = stateless_client
        self.cmd_name = kwargs.get('prog')
        self.register('action', 'merge', _MergeAction)

    def add_arg_list(self, *args):
        populate_parser(self, *args)

    # a simple hook for add subparsers to add stateless client

    def add_subparsers(self, *args, **kwargs):
        sub = super(CCmdArgParser, self).add_subparsers(*args, **kwargs)

        # save pointer to the original add parser method
        add_parser = sub.add_parser
        stateless_client = self.stateless_client

        def add_parser_hook(self, *args, **kwargs):
            parser = add_parser(self, *args, **kwargs)
            parser.stateless_client = stateless_client
            return parser

        # override with the hook
        sub.add_parser = add_parser_hook

        def remove_parser(name):
            if name in sub._name_parser_map:
                del sub._name_parser_map[name]
                for action in sub._choices_actions:
                    if action.dest == name:
                        sub._choices_actions.remove(action)
            else:
                self._print_message(
                    bold('Subparser "%s" does not exist!' % name))

        sub.remove_parser = remove_parser
        sub.has_parser = lambda name: name in sub._name_parser_map
        sub.get_parser = lambda name: sub._name_parser_map.get(name)

        return sub

    # hook this to the logger

    def _print_message(self, message, file=None):
        self.stateless_client.logger.log(message)

    def error(self, message):
        self.print_usage()
        self._print_message(('%s: error: %s\n') % (self.prog, message))
        raise ValueError(message)

    def has_ports_cfg(self, opts):
        return hasattr(opts, "all_ports") or hasattr(opts, "ports")

    def parse_args(self, args=None, namespace=None, default_ports=None, verify_acquired=False):
        try:
            opts = super(CCmdArgParser, self).parse_args(args, namespace)
            if opts is None:
                return RC_ERR("'{0}' - invalid arguments".format(self.cmd_name))

            if not self.has_ports_cfg(opts):
                return opts
            opts.ports = listify(opts.ports)

            # if all ports are marked or
            if(getattr(opts, "all_ports", None) is True) or(getattr(opts, "ports", None) == []):
                if default_ports is None:
                    opts.ports = self.stateless_client.get_acquired_ports()
                else:
                    opts.ports = default_ports

            opts.ports = list_remove_dup(opts.ports)

            # so maybe we have ports configured
            invalid_ports = list_difference(
                opts.ports, self.stateless_client.get_all_ports())
            if invalid_ports:

                if len(invalid_ports) > 1:
                    msg = "{0}: port(s) {1} are not valid port IDs".format(
                        self.cmd_name, invalid_ports)
                else:
                    msg = "{0}: port {1} is not a valid port ID".format(
                        self.cmd_name, invalid_ports[0])

                self.stateless_client.logger.log(format_text(msg, 'bold'))
                return RC_ERR(msg)

            # verify acquired ports
            if verify_acquired:
                acquired_ports = self.stateless_client.get_acquired_ports()

                diff = list_difference(opts.ports, acquired_ports)
                if diff:
                    msg = "{0} - port(s) {1} are not acquired".format(self.cmd_name, diff)
                    self.stateless_client.logger.log(format_text(msg, 'bold'))
                    return RC_ERR(msg)

                # no acquire ports at all
                if not acquired_ports:
                    msg = "{0} - no acquired ports".format(self.cmd_name)
                    self.stateless_client.logger.log(format_text(msg, 'bold'))
                    return RC_ERR(msg)

            return opts

        except ValueError as e:
            return RC_ERR("'{0}' - {1}".format(self.cmd_name, str(e)))

        except SystemExit:
            # recover from system exit scenarios, such as "help", or bad arguments.
            return RC_ERR("'{0}' - {1}".format(self.cmd_name, "no action"))

    def formatted_error(self, msg):
        self.print_usage()
        self._print_message(('%s: error: %s\n') % (self.prog, msg))


def get_flags(opt):
    return OPTIONS_DB[opt].name_or_flags


def populate_parser(parser, *args):
    for param in args:
        try:

            if isinstance(param, int):
                argument = OPTIONS_DB[param]
            else:
                argument = param

            if isinstance(argument, ArgumentGroup):
                if argument.type == MUTEX:
                    # handle as mutually exclusive group
                    group = parser.add_mutually_exclusive_group(
                        **argument.options)
                    for sub_argument in argument.args:
                        group.add_argument(*OPTIONS_DB[sub_argument].name_or_flags,
                                           **OPTIONS_DB[sub_argument].options)

                elif argument.type == NON_MUTEX:
                    group = parser.add_argument_group(**argument.options)
                    for sub_argument in argument.args:
                        group.add_argument(*OPTIONS_DB[sub_argument].name_or_flags,
                                           **OPTIONS_DB[sub_argument].options)
                else:
                    raise Exception(
                        'Invalid ArgumentGroup type, should be either MUTEX or NON_MUTEX')
            elif isinstance(argument, ArgumentPack):
                parser.add_argument(*argument.name_or_flags,
                                    **argument.options)
            else:
                raise Exception(
                    'Invalid arg object, should be ArgumentGroup or ArgumentPack, got: %s'
                    % type(argument))
        except KeyError as e:
            cause = e.args[0]
            raise KeyError(
                "The attribute '{0}' is missing as a field of the {1} option.\n"
                .format(cause, param))


def gen_parser(stateless_client, op_name, description, *args, **kw):
    parser = CCmdArgParser(stateless_client, prog=op_name, conflict_handler='resolve',
                           description=description, **kw)

    populate_parser(parser, *args)
    return parser


if __name__ == "__main__":
    pass



#!/usr/bin/python3

import os
# import warnings
import platform
import sys

# if not set - set it to default
# TREX_STL_EXT_PATH = os.environ.get('TREX_STL_EXT_PATH')

# take default
# if not TREX_STL_EXT_PATH or not os.path.exists(TREX_STL_EXT_PATH):
#     CURRENT_PATH = os.path.dirname(os.path.realpath(__file__))
#     TREX_STL_EXT_PATH = os.path.normpath(os.path.join(
#         CURRENT_PATH, os.pardir, os.pardir, 'external_libs'))
# if not os.path.exists(TREX_STL_EXT_PATH):
# ../../../../external_libs
#     TREX_STL_EXT_PATH = os.path.normpath(os.path.join(
#         CURRENT_PATH, os.pardir, os.pardir, os.pardir, os.pardir,
#         'external_libs'))
# if not os.path.exists(TREX_STL_EXT_PATH):
#     raise Exception(
#         'Could not determine path of external_libs,'
#         + ' try setting TREX_STL_EXT_PATH variable')

# the modules required
# py-dep requires python2/python3 directories
# arch-dep requires ucs2/ucs4 and 32bit/64bit directories
CLIENT_UTILS_MODULES = [{'name': 'texttable-0.8.4'},
                        {'name': 'pyyaml-3.11', 'py-dep': True},
                        {'name': 'scapy-2.3.1', 'py-dep': True},
                        {'name': 'pyzmq-14.5.0', 'py-dep': True,
                        'arch-dep': True},
                        {'name': 'simpy-3.0.10'},
                        {'name': 'trex-openssl'},
                        ]


def generate_module_path(module, is_python3, is_64bit, is_ucs2):
    platform_path = [module['name']]

    if module.get('py-dep'):
        platform_path.append('python3' if is_python3 else 'python2')

    if module.get('arch-dep'):
        platform_path.append('ucs2' if is_ucs2 else 'ucs4')
        platform_path.append('64bit' if is_64bit else '32bit')

    return os.path.normcase(os.path.join(TREX_STL_EXT_PATH, *platform_path))


def import_module_list(modules_list):

    # platform data
    is_64bit = platform.architecture()[0] == '64bit'
    is_python3 = (sys.version_info >= (3, 0))
    is_ucs2 = (sys.maxunicode == 65535)

    # regular modules
    for p in modules_list:
        full_path = generate_module_path(p, is_python3, is_64bit, is_ucs2)

        if not os.path.exists(full_path):
            print(
                ("Unable to find required module library: '{0}'"
                 .format(p['name'])))
            print("Please provide the correct path"
                  + " using TREX_STL_EXT_PATH variable")
            print(("current path used: '{0}'".format(full_path)))
            exit(1)

        sys.path.insert(1, full_path)


import_module_list(CLIENT_UTILS_MODULES)



#!/usr/bin/python3
"""
Manage port state validation

Author:
  Itay Marom

"""

from trex_stl_lib.trex_stl_exceptions import STLError

from trex_stl_lib.utils.common import list_difference, list_remove_dup

# type of states to be enforced
# types starting with _ are internally used
_PSV_ALL, PSV_UP, PSV_ACQUIRED, PSV_IDLE, PSV_RESOLVED, PSV_SERVICE, PSV_NON_SERVICE, PSV_L3 = list(
    range(0, 8))


class PortState(object):
    '''
        abstract class to create a state validation
    '''

    def validate(self, client, cmd_name, ports, custom_err_msg=None):
        invalid_ports = list_difference(ports, self.get_valid_ports(client))
        if invalid_ports:
            err_msg = self.def_err_msg()
            if custom_err_msg:
                err_msg = '{0} - {1}'.format(err_msg, custom_err_msg)

            raise STLError(
                '{0} - port(s) {1}: {2}'.format(cmd_name, invalid_ports, err_msg))

    def get_valid_ports(self, client):
        raise NotImplementedError


class PortStateAll(PortState):
    def def_err_msg(self):
        return 'invalid port IDs'

    def get_valid_ports(self, client):
        return client.get_all_ports()


class PortStateUp(PortState):
    def def_err_msg(self):
        return 'link is DOWN'

    def get_valid_ports(self, client):
        return [port_id for port_id in client.get_all_ports() if client.ports[port_id].is_up()]


class PortStateAcquired(PortState):
    def def_err_msg(self):
        return 'must be acquired'

    def get_valid_ports(self, client):
        return client.get_acquired_ports()


class PortStateIdle(PortState):
    def def_err_msg(self):
        return 'are active'

    def get_valid_ports(self, client):
        return [port_id for port_id in client.get_all_ports()
                if not client.ports[port_id].is_active()]


class PortStateService(PortState):
    def def_err_msg(self):
        return 'must be under service mode'

    def get_valid_ports(self, client):
        return client.get_service_enabled_ports()


class PortStateNonService(PortState):
    def def_err_msg(self):
        return 'cannot be under service mode'

    def get_valid_ports(self, client):
        return [port_id for port_id in client.get_all_ports()
                if not client.ports[port_id].is_service_mode_on()]


class PortStateL3(PortState):
    def def_err_msg(self):
        return 'does not have a valid IPv4 address'

    def get_valid_ports(self, client):
        return [port_id for port_id in client.get_all_ports() if client.ports[port_id].is_l3_mode()]


class PortStateResolved(PortState):
    def def_err_msg(self):
        return 'must have resolved destination address'

    def get_valid_ports(self, client):
        return client.get_resolved_ports()


class PortStateValidator(object):
    '''
        port state validator

        used to validate different groups of states
        required for 'ports'
    '''

    def __init__(self, client):
        self.client = client

        self.validators = {}

        self.validators[_PSV_ALL] = PortStateAll()
        self.validators[PSV_UP] = PortStateUp()
        self.validators[PSV_ACQUIRED] = PortStateAcquired()
        self.validators[PSV_IDLE] = PortStateIdle()
        self.validators[PSV_RESOLVED] = PortStateResolved()
        self.validators[PSV_SERVICE] = PortStateService()
        self.validators[PSV_L3] = PortStateL3()
        self.validators[PSV_NON_SERVICE] = PortStateNonService()

    def validate(self, cmd_name, ports, states=None, allow_empty=False):
        '''
            main validator

        '''

        # listify
        ports = ports if isinstance(ports, (list, tuple)) else [ports]
        ports = list_remove_dup(ports)

        if not ports and not allow_empty:
            raise STLError(
                '{0} - action requires at least one port'.format(cmd_name))

        # default checks for every command
        states_map = {_PSV_ALL: None}

        # eventually states is a map from every mandatory state to it's error message
        if isinstance(states, int):
            states_map[states] = None
        elif isinstance(states, (list, tuple)):
            for s in states:
                states_map[s] = None
        elif isinstance(states, dict):
            states_map.update(states)
        else:
            pass

        for state, err_msg in list(states_map.items()):
            self.validators[state].validate(
                self.client, cmd_name, ports, err_msg)

        # returns the ports listify
        return ports



#!/usr/bin/python3

"""
DHCP service implementation

Description:
    <FILL ME HERE>

How to use:
    <FILL ME HERE>

Author:
  Itay Marom

"""
from .trex_stl_service import STLService, STLServiceFilter
from .trex_stl_dhcp_parser import DHCPParser

from trex_stl_lib.trex_stl_exceptions import STLError

from scapy.layers.dhcp import DHCP, BOOTP
from scapy.layers.l2 import Ether
from scapy.layers.inet import IP, UDP

from collections import defaultdict
import random
import struct
import socket
import re

# create a single global parser
parser = DHCPParser()


def ipv4_num_to_str(num):
    return socket.inet_ntoa(struct.pack('!I', num))


def ipv4_str_to_num(ipv4_str):
    return struct.unpack("!I", socket.inet_aton(ipv4_str))[0]


class STLServiceFilterDHCP(STLServiceFilter):
    '''
        Service filter for DHCP services
    '''

    def __init__(self):
        self.services = defaultdict(list)

    def add(self, service):
        self.services[service.get_xid()].append(service)

    def lookup(self, pkt):
        # correct XID is enough to verify ownership
        xid = parser.parse(pkt).xid

        return self.services.get(xid, [])

    def get_bpf_filter(self):
        return 'udp port 67 or 68'


################### internal ###################
class STLServiceDHCP(STLService):

    # DHCP states
    INIT, SELECTING, REQUESTING, BOUND = list(range(4))

    def __init__(self, mac, verbose_level=STLService.ERROR):

        # init the base object
        super(STLServiceDHCP, self).__init__(verbose_level)

        self.xid = random.getrandbits(32)

        self.mac = mac
        self.mac_bytes = self.mac2bytes(mac)

        self.record = None
        self.state = 'INIT'

    def get_filter_type(self):
        return STLServiceFilterDHCP

    def get_xid(self):
        return self.xid

    def get_mac(self):
        return self.mac

    def mac2bytes(self, mac):
        if type(mac) != str or not re.match("[0-9a-f]{2}([:])[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$", mac.lower()):
            self.err('invalid MAC format: {}'.format(mac))

        return struct.pack('B' * 6, *[int(b, 16) for b in mac.split(':')])

    #########################  protocol state machines  #########################

    def run(self, pipe):

        # while running under 'INIT' - perform acquire
        if self.state == 'INIT':
            return self._acquire(pipe)
        elif self.state == 'BOUND':
            return self._release(pipe)

    def _acquire(self, pipe):
        '''
            Acquire DHCP lease protocol
        '''

        # main state machine loop
        self.state = 'INIT'
        self.record = None
        self.retries = 5

        while True:

            # INIT state
            if self.state == 'INIT':

                self.retries -= 1
                if self.retries <= 0:
                    break

                self.log('DHCP: {0} ---> DISCOVERY'.format(self.mac))

                # send a discover message
                yield pipe.async_tx_pkt(parser.disc(self.xid, self.mac_bytes))

                self.state = 'SELECTING'
                continue

            # SELECTING state
            elif self.state == 'SELECTING':

                # wait until packet arrives or timeout occurs
                pkts = yield pipe.async_wait_for_pkt(3)
                pkts = [pkt['pkt'] for pkt in pkts]

                # filter out the offer responses
                offers = [parser.parse(pkt) for pkt in pkts]
                offers = [
                    offer for offer in offers if offer.options['message-type'] == parser.OFFER]

                if not offers:
                    self.log('DHCP: {0} *** timeout on offers - retries left: {1}'.format(
                        self.mac, self.retries), level=STLService.ERROR)
                    self.state = 'INIT'
                    continue

                offer = offers[0]
                self.log("DHCP: {0} <--- OFFER from '{1}' with address '{2}' ".format(
                    self.mac, ipv4_num_to_str(offer.options['server_id']), ipv4_num_to_str(offer.yiaddr)))

                self.state = 'REQUESTING'
                continue

            # REQUEST state
            elif self.state == 'REQUESTING':
                self.retries = 5

                self.log('DHCP: {0} ---> REQUESTING'.format(self.mac))

                # send the request
                yield pipe.async_tx_pkt(parser.req(self.xid, self.mac_bytes, offer.yiaddr))

                # wait for response
                pkts = yield pipe.async_wait_for_pkt(3)
                pkts = [pkt['pkt'] for pkt in pkts]

                # filter out the offer responses
                acknacks = [parser.parse(pkt) for pkt in pkts]
                acknacks = [
                    acknack for acknack in acknacks if acknack.options['message-type'] in(parser.ACK, parser.NACK)]

                if not acknacks:
                    self.log('DHCP: {0} *** timeout on ack - retries left: {1}'.format(
                        self.mac, self.retries), level=STLService.ERROR)
                    self.state = 'INIT'
                    continue

                # by default we choose the first one... usually there should be only one response
                acknack = acknacks[0]

                if acknack.options['message-type'] == parser.ACK:
                    self.log("DHCP: {0} <--- ACK from '{1}' to address '{2}' ".format(
                        self.mac, ipv4_num_to_str(offer.options['server_id']), ipv4_num_to_str(offer.yiaddr)))
                    self.state = 'BOUND'
                else:
                    self.log("DHCP: {0} <--- NACK from '{1}'".format(self.mac,
                                                                     ipv4_num_to_str(offer.options['server_ip'])))
                    self.state = 'INIT'

                continue

            elif self.state == 'BOUND':

                # parse the offer and save it
                self.record = self.DHCPRecord(offer)
                break

    def _release(self, pipe):
        '''
            Release the DHCP lease
        '''
        self.log('DHCP: {0} ---> RELEASING'.format(self.mac))

        release_pkt = parser.release(self.xid,
                                     self.record.client_mac,
                                     ipv4_str_to_num(self.record.client_ip),
                                     self.record.server_mac,
                                     ipv4_str_to_num(self.record.server_ip))

        yield pipe.async_tx_pkt(release_pkt)

        # clear the record
        self.record = None

    def get_record(self):
        '''
            Returns a DHCP record
        '''
        return self.record

    class DHCPRecord(object):

        def __init__(self, offer):

            self.server_mac = offer.srcmac
            self.client_mac = offer.dstmac

            options = offer.options

            self.server_ip = ipv4_num_to_str(
                options['server_id']) if 'server_id' in options else 'N/A'
            self.subnet = ipv4_num_to_str(
                options['subnet_mask']) if 'subnet_mask' in options else 'N/A'
            self.client_ip = ipv4_num_to_str(offer.yiaddr)

            self.domain = options.get('domain', 'N/A')
            self.lease = options.get('lease-time', 'N/A')

        def __str__(self):
            return "ip: {0}, server_ip: {1}, subnet: {2}, domain: {3}, lease_time: {4}".format(self.client_ip, self.server_ip, self.subnet, self.domain, self.lease)
